"""
field_band_pass_fc
==================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.Native plugin, from "filter" category
"""

class field_band_pass_fc(Operator):
    """The band pass filter returns all the values strictly superior to the min threshold value and stricly inferior to the max threshold value in input.

      available inputs:
        - fields_container (FieldsContainer)
        - min_threshold (float, Field)
        - max_threshold (float, Field)

      available outputs:
        - fields_container (FieldsContainer)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.filter.field_band_pass_fc()

      >>> # Make input connections
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_min_threshold = float()
      >>> op.inputs.min_threshold.connect(my_min_threshold)
      >>> my_max_threshold = float()
      >>> op.inputs.max_threshold.connect(my_max_threshold)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.filter.field_band_pass_fc(fields_container=my_fields_container,min_threshold=my_min_threshold,max_threshold=my_max_threshold)

      >>> # Get output data
      >>> result_fields_container = op.outputs.fields_container()"""
    def __init__(self, fields_container=None, min_threshold=None, max_threshold=None, config=None, server=None):
        super().__init__(name="core::field::band_pass_fc", config = config, server = server)
        self._inputs = InputsFieldBandPassFc(self)
        self._outputs = OutputsFieldBandPassFc(self)
        if fields_container !=None:
            self.inputs.fields_container.connect(fields_container)
        if min_threshold !=None:
            self.inputs.min_threshold.connect(min_threshold)
        if max_threshold !=None:
            self.inputs.max_threshold.connect(max_threshold)

    @staticmethod
    def _spec():
        spec = Specification(description="""The band pass filter returns all the values strictly superior to the min threshold value and stricly inferior to the max threshold value in input.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""field or fields container with only one field is expected"""), 
                                 1 : PinSpecification(name = "min_threshold", type_names=["double","field"], optional=False, document="""a min threshold scalar or a field containing one value is expected"""), 
                                 2 : PinSpecification(name = "max_threshold", type_names=["double","field"], optional=False, document="""a max threshold scalar or a field containing one value is expected""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "fields_container", type_names=["fields_container"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "core::field::band_pass_fc")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsFieldBandPassFc 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsFieldBandPassFc 
        """
        return super().outputs


#internal name: core::field::band_pass_fc
#scripting name: field_band_pass_fc
class InputsFieldBandPassFc(_Inputs):
    """Intermediate class used to connect user inputs to field_band_pass_fc operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.filter.field_band_pass_fc()
      >>> my_fields_container = dpf.FieldsContainer()
      >>> op.inputs.fields_container.connect(my_fields_container)
      >>> my_min_threshold = float()
      >>> op.inputs.min_threshold.connect(my_min_threshold)
      >>> my_max_threshold = float()
      >>> op.inputs.max_threshold.connect(my_max_threshold)
    """
    def __init__(self, op: Operator):
        super().__init__(field_band_pass_fc._spec().inputs, op)
        self._fields_container = Input(field_band_pass_fc._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._fields_container)
        self._min_threshold = Input(field_band_pass_fc._spec().input_pin(1), 1, op, -1) 
        self._inputs.append(self._min_threshold)
        self._max_threshold = Input(field_band_pass_fc._spec().input_pin(2), 2, op, -1) 
        self._inputs.append(self._max_threshold)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator

        - pindoc: field or fields container with only one field is expected

        Parameters
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.filter.field_band_pass_fc()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> #or
        >>> op.inputs.fields_container(my_fields_container)

        """
        return self._fields_container

    @property
    def min_threshold(self):
        """Allows to connect min_threshold input to the operator

        - pindoc: a min threshold scalar or a field containing one value is expected

        Parameters
        ----------
        my_min_threshold : float, Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.filter.field_band_pass_fc()
        >>> op.inputs.min_threshold.connect(my_min_threshold)
        >>> #or
        >>> op.inputs.min_threshold(my_min_threshold)

        """
        return self._min_threshold

    @property
    def max_threshold(self):
        """Allows to connect max_threshold input to the operator

        - pindoc: a max threshold scalar or a field containing one value is expected

        Parameters
        ----------
        my_max_threshold : float, Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.filter.field_band_pass_fc()
        >>> op.inputs.max_threshold.connect(my_max_threshold)
        >>> #or
        >>> op.inputs.max_threshold(my_max_threshold)

        """
        return self._max_threshold

class OutputsFieldBandPassFc(_Outputs):
    """Intermediate class used to get outputs from field_band_pass_fc operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.filter.field_band_pass_fc()
      >>> # Connect inputs : op.inputs. ...
      >>> result_fields_container = op.outputs.fields_container()
    """
    def __init__(self, op: Operator):
        super().__init__(field_band_pass_fc._spec().outputs, op)
        self._fields_container = Output(field_band_pass_fc._spec().output_pin(0), 0, op) 
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator


        Returns
        ----------
        my_fields_container : FieldsContainer, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.filter.field_band_pass_fc()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container() 
        """
        return self._fields_container




"""
.. _ref_data_sources:
    
DataSources
===========
"""
import os

from ansys import dpf
from ansys.grpc.dpf import data_sources_pb2, data_sources_pb2_grpc, base_pb2
from ansys.dpf.core.errors import protect_grpc


class DataSources:
    """The data sources is a container of files on which the analysis results can be found.
     An extension key ('rst' for example) is used to choose which files represent results files,
     the other one being accessory files.
     A result file path can be set in the DataSOurces initializer.
   
    Parameters
    ----------
    result_path : str, optional
        path of the result

    data_sources : ansys.grpc.dpf.data_sources_pb2.DataSources
        gRPC data sources message.

    server : DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    Initialize a model from a result path
    >>> from ansys.dpf import core as dpf
    >>> my_data_sources = dpf.DataSources('file.rst')
    >>> my_data_sources.result_files
    ['file.rst']
    
    """

    def __init__(self, result_path=None, data_sources=None, server=None):
        """Initialize connection with the server"""
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect()

        if data_sources is None:
            request = base_pb2.Empty()
            self._message = self._stub.Create(request)
        else:
            self._message = data_sources

        if result_path is not None:
            self.set_result_file_path(result_path)

    @protect_grpc
    def _connect(self):
        """Connect to the grpc service"""
        return data_sources_pb2_grpc.DataSourcesServiceStub(self._server.channel)

    def set_result_file_path(self, filepath, key=""):
        """Add a result file path to the datasources. 
        The result file path key is used to choose the right plugin 
        when a result is asked with an Operator. 

        Parameters
        ----------
        filepath : str
            Path to the result file.

        key : str, optional
            Extension of the file, found directly if it is not set.

        Examples
        --------
        Create a data source and set the result file path

        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_result_file_path('/tmp/file.rst')
        >>> data_sources.result_files
        ['/tmp/file.rst']
        
        """
        request = data_sources_pb2.UpdateRequest()
        request.result_path = True
        request.key = key
        request.path = filepath
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)
        
    def set_domain_result_file_path(self, path, domain_id):
        """Add a result file path by domain. This method can be called to 
        to handle files created by a distributed solve.
        
        Parameters
        ----------
        path: str
            Path to the file.
                
        domain_id: int, optional, default is 0
            Domain id for distributed files.
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.set_domain_result_file_path('/tmp/file0.sub', 0)
        >>> data_sources.set_domain_result_file_path('/tmp/file1.sub', 1)
        
        """
        request = data_sources_pb2.UpdateRequest()
        request.result_path = True
        request.domain.domain_path = True
        request.domain.domain_id = domain_id
        request.path = path
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)  

    def add_file_path(self, filepath, key="", is_domain: bool = False, domain_id = 0):
        """Add a file path to the data sources. The files not added as
        result files are accessory files used to access accessory information
        not present in the result files.

        Parameters
        ----------
        filepath : str
            Path of the file.
            
        key : str, optional
            Extension of the file, found directly if it is not set.
            
        is_domain: bool
            file_path is domain_path

        domain_id: int, optional, default is 0
            Domain id for distributed files.
            Must be set with domain_path = True to be taken into account. 

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> data_sources = dpf.DataSources()
        >>> data_sources.add_file_path('/tmp/ds.dat')
        
        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))

        request = data_sources_pb2.UpdateRequest()
        request.key = key
        request.path = filepath
        if is_domain:
            request.domain.domain_path = True
            request.domain.domain_id = domain_id
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)
        
    def add_file_path_for_specified_result(self, filepath, key="", result_key = ""):
        """Add a file path to the data sources. The files not added as
        result files are accessory files used to access accessory information
        not present in the result files. Using the ``'specified_result'`` option
        is used when several results files with different results keys are used 
        in the data sources. For example, if a solve used 2 different solvers
        which generated 2 different sets of files, this method can be used.

        Parameters
        ----------
        filepath : str
            Path of the file.
            
        key : str, optional
            Extension of the file, found directly if it is not set.
            
        result_key: str, optional
            Extension of the result file with which this file path belongs.
        """
        # The filename needs to be a fully qualified file name
        if not os.path.dirname(filepath):
            # append local path
            filepath = os.path.join(os.getcwd(), os.path.basename(filepath))

        request = data_sources_pb2.UpdateRequest()
        request.key = key
        request.result_key = result_key
        request.path = filepath
        request.data_sources.CopyFrom(self._message)
        self._stub.Update(request)

    def add_upstream(self, upstream_data_sources):
        """Add an upstream datasources.

        This is used to add a set of path creating an upstream for
        recursive workflows.

        Parameters
        ----------
        upstream_data_sources : DataSources

        """
        request = data_sources_pb2.UpdateUpstreamRequest()
        request.upstream_data_sources.CopyFrom(upstream_data_sources._message)
        request.data_sources.CopyFrom(self._message)
        self._stub.UpdateUpstream(request)

    @property
    def result_key(self):
        """Returns the result key used by the data sources
        
        Returns
        ----------
        result_key : str
        """
        return self._info["result_key"]

    @property
    def result_files(self):
        """Returns the list of result files contained
        by the data sources
        
        Returns
        ----------
        result_files : list of str
        """
        key = self.result_key
        if (key == ''): 
            return None
        else: 
            return self._info["paths"][key]

    @property
    def _info(self):
        list = self._stub.List(self._message)
        paths = {}
        for key in list.paths:
            key_paths=[]
            for path in list.paths[key].paths:
                key_paths.append(path)
            paths[key] = key_paths
        out = {"result_key": list.result_key, "paths": paths}
        return out

    def __str__(self):
        """Describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)

    def __del__(self):
        try:  # should silently fail
            self._stub.Delete(self._message)
        except:
            pass



"""
Server
======
Contains the directives necessary to start the dpf server.
"""
from threading import Thread
import io
import platform
import logging
import time
import os
import socket
import subprocess
import grpc
import psutil
import weakref
import atexit
import copy

from ansys import dpf
from ansys.dpf.core.misc import find_ansys, is_ubuntu
from ansys.dpf.core import errors

from ansys.dpf.core._version import __ansys_version__

MAX_PORT = 65535

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')

# default DPF server port
DPF_DEFAULT_PORT =  int(os.environ.get('DPF_PORT',50054))
LOCALHOST = os.environ.get('DPF_IP','127.0.0.1')

def shutdown_global_server():
    try :
        if dpf.core.SERVER != None:
            del dpf.core.SERVER
    except:
        pass
        
atexit.register(shutdown_global_server)

def has_local_server():
    """Returns True when a local DPF gRPC server has been created"""
    return dpf.core.SERVER is not None

def _global_server():
    """Return the global server if it exists.

    If the global server has not been specified, check if the user
    has specified the "DPF_START_SERVER" environment variable.  If
    ``True``, start the server locally.  If ``False``, connect to the
    existing server.
    """
    if dpf.core.SERVER is None:
        if os.environ.get('DPF_START_SERVER', '').lower() == 'false':
            ip = os.environ.get('DPF_IP', LOCALHOST)
            port = int(os.environ.get('DPF_PORT', DPF_DEFAULT_PORT))
            connect_to_server(ip, port)
        else:
            start_local_server()
            
    return dpf.core.SERVER


def port_in_use(port, host=LOCALHOST):
    """Returns True when a port is in use at the given host.

    Must actually "bind" the address.  Just checking if we can create
    a socket is insufficient as it's possible to run into permission
    errors like:

    - An attempt was made to access a socket in a way forbidden by its
      access permissions.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        try:
            sock.bind((host, port))
            return False
        except:
            return True


def check_valid_ip(ip):
    """Raises an error when an invalid ip is entered"""
    try:
        socket.inet_aton(ip)
    except OSError:
        raise ValueError(f'Invalid IP address "{ip}"')


def shutdown_all_session_servers():
    """Close all active servers created by this module"""
    from ansys.dpf.core import _server_instances
    copy_instances = copy.deepcopy(_server_instances)
    for instance in copy_instances:
        try:
            instance().shutdown()
        except Exception as e:
            print(e.args)
            pass


def start_local_server(ip=LOCALHOST, port=DPF_DEFAULT_PORT,
                       ansys_path=None, as_global=True, load_operators=True):
    """Starts a new local DPF server at a given port and ip.
    Requires Windows and ANSYS v211 or newer be installed.
    If as_global is set to True (default) then the server is stored in the module
    (replacing the one stored before). Else, the user must keep a handle on 
    his/her server

    Parameters
    ----------
    ip : str
        IP address of the remote or local instance to connect to.

    port : int
        Port to connect to the remote instance on.

    ansys_path : str, optional
        Root path containing ansys.  For example ``/ansys_inc/v212/``.
        Defaults to the latest install of ANSYS.

    as_global : bool, optional
        Stores this ip and port as global variables for the dpf
        module.  All DPF objects created in this Python session will
        use this IP and port.  Default ``True``.
        
    load_operators : bool, optional
        Automatically load the math operators

    Returns
    -------
    server : server.DpfServer
    """
    if ansys_path is None:
        ansys_path = os.environ.get('AWP_ROOT'+__ansys_version__, find_ansys())
    if ansys_path is None:
        raise ValueError('Unable to automatically locate the Ansys path.  '
                         'Manually enter one when starting the server or set it '
                         'as the environment variable "ANSYS_PATH"')

    # verify path exists
    if not os.path.isdir(ansys_path):
        raise NotADirectoryError(f'Invalid Ansys path "{ansys_path}"')

    # parse the version to an int and check for supported
    try:
        ver = int(ansys_path[-3:])
        if ver < 211:
            raise errors.InvalidANSYSVersionError(f'Ansys v{ver} does not support DPF')
        if ver == 211 and is_ubuntu():
            raise OSError('DPF on v211 does not support Ubuntu')
    except ValueError:
        pass
    

    # avoid using any ports in use from existing servers    
    used_ports=[]
    if dpf.core._server_instances:
        for srv in dpf.core._server_instances:
            if srv():
                used_ports.append(srv().port)

    while port in used_ports:
        port += 1

    # verify port is free
    while port_in_use(port):
        port += 1

    server = None
    n_attempts = 10
    for _ in range(n_attempts):
        try:
            server = DpfServer(ansys_path, ip, port,as_global= as_global, load_operators = load_operators)
            break
        except errors.InvalidPortError:  # allow socket in use errors
            port += 1

    if server is None:
        raise OSError(f'Unable to launch the server after {n_attempts} attempts.  '
                      'Check the following path:\n{ansys_path}\n\n'
                      'or attempt to use a different port')

    dpf.core._server_instances.append(weakref.ref(server))
    return server


def connect_to_server(ip=LOCALHOST, port=DPF_DEFAULT_PORT, as_global=True, timeout=5):
    """Connect to an existing dpf server.

    Set this as the global default channel that will be used for the
    duration of dpf.
    
    Parameters
    ----------
    ip : str
        IP address of the remote or local instance to connect to.

    port : int
        Port to connect to the remote instance on.
        
    as_global : bool, optional
        Stores this ip and port as global variables for the dpf
        module.  All DPF objects created in this Python session will
        use this IP and port.  Default ``True``.

    timeout : float
        Maximum timeout to connect to the DPF server.

    Examples
    --------
    
    >>> from ansys.dpf import core as dpf
        
    Create a server
    
    >>> #server = dpf.start_local_server(ip = '127.0.0.1')
    >>> #port = server.port
    
    Connect to a remote server at a non-default port

    >>> #specified_server = dpf.connect_to_server('127.0.0.1', port, as_global=False)

    Connect to the localhost at the default port

    >>> #unspecified_server = dpf.connect_to_server(as_global=False)
    
    """
    server = DpfServer(ip=ip, port=port, as_global=as_global, launch_server=False)
    dpf.core._server_instances.append(weakref.ref(server))
    return server


class DpfServer:
    """Instance of the DPF server
    
    Parameters
    -----------
    server_bin : str
        Location of the dpf executable.

    ip : str
        IP address of the remote or local instance to connect to.

    port : int
        Port to connect to the remote instance on.  Defaults to 50054.

    timeout : float, optional
        Fails when a connection takes longer than ``timeout`` seconds
        to initialize.

    as_global : bool, optional
        Stores this ip and port as global variables for the dpf
        module.  All DPF objects created in this Python session will
        use this IP and port.

    load_operators : bool, optional
        Automatically load the math operators

    ansys_path : str, optional
        Path containing ansys.  For example ``/ansys_inc/v212/``
        
   launch_server : bool, optional
        Launch the server on windows
    """

    def __init__(self, ansys_path="", ip=LOCALHOST, port=DPF_DEFAULT_PORT,
                 timeout=10, as_global=True, load_operators=True, launch_server=True):
        """Start the dpf server server"""
        # check valid ip and port
        check_valid_ip(ip)
        if not isinstance(port, int):
            raise ValueError('Port must be an integer')

        if os.name == 'posix' and 'ubuntu' in platform.platform().lower():
            raise OSError('DPF does not support Ubuntu')
        elif launch_server:
            launch_dpf(ansys_path, ip, port)        

        self.channel = grpc.insecure_channel('%s:%d' % (ip, port))
        
        if launch_server is False:
            state = grpc.channel_ready_future(self.channel)
            # verify connection has matured
            tstart = time.time()
            while ((time.time() - tstart) < timeout) and not state._matured:
                time.sleep(0.01)
        
            if not state._matured:
                raise TimeoutError(f'Failed to connect to {ip}:{port} in {timeout} seconds')
        
            LOG.debug('Established connection to DPF gRPC')
        
        # assign to global channel when requested
        if as_global:
            dpf.core.SERVER =self

        # TODO: add to PIDs ...

        # store port and ip for later reference
        self.live = True
        self.ansys_path=ansys_path
        self._input_ip = ip
        self._input_port = port
        self._own_process = launch_server
        
    @property
    def _base_service(self):
        if not hasattr(self,"__base_service"):            
            from ansys.dpf.core.core import BaseService
            self.__base_service = BaseService(self, timeout=1)
        return self.__base_service
    
    @property
    def info(self):
        """Recover server information
           
           Returns
           -------
           info : dictionary
               dictionary with "server_ip", "server_port", "server_process_id"
               "server_version" keys
        """
        return self._base_service.server_info
    
    @property
    def ip(self):
        """Get the ip of the server
        
        Returns
        -------
        ip : str
        """
        try:
            return self._base_service.server_info["server_ip"]
        except:
            return ""
    
    @property
    def port(self):
        """Get the port of the server
        
        Returns
        -------
        port : int
        """
        try:
            return self._base_service.server_info["server_port"]
        except:
            return 0
    
    @property
    def version(self):
        """Get the version of the server
        
        Returns
        -------
        version : str
        """
        return self._base_service.server_info["server_version"]

    def __str__(self): 
        return f'DPF Server: {self.info}'

    def shutdown(self):
        if self._own_process and self.live and self._base_service:
            self._base_service._prepare_shutdown()
            p = psutil.Process(self._base_service.server_info["server_process_id"])
            p.kill()
            time.sleep(0.1)
            self.live = False
            try:
                if id(dpf.core.SERVER) == id(self):
                    dpf.core.SERVER =None
            except:
                pass
                
            try:
                for i, server in enumerate(dpf.core._server_instances):
                    if server() == self:
                        dpf.core._server_instances.remove(server)
            except:
                pass
                     
    def __eq__(self,other_server):
        """Return true, if the ip and the port are equals"""
        if isinstance(other_server, DpfServer):
            return self.ip == other_server.ip and self.port == other_server.port
        return False
    
    def __ne__(self,other_server):
        """Return true, if the ip or the port are different"""
        return not self.__eq__(other_server)
    

    
    def __del__(self):
        try:
            self.shutdown()
        except:
            pass
        
    def check_version(self, required_version, msg = None):
        """
        Check if the server version matches with a required version.
        
        Parameters
        ----------
        required_version : str
            Required version that will be compared with the server version.
        msg : str, optional
            Message to be contained in the raised Exception if versions are
            not meeting.
    
        Raises
        ------
        dpf_errors : errors
            errors.DpfVersionNotSupported is raised if failure.
    
        Returns
        -------
        bool : 
            True if the server version meets the requirement.
        """
        from ansys.dpf.core.check_version import server_meet_version_and_raise
        return server_meet_version_and_raise(required_version, self, msg)



def launch_dpf(ansys_path, ip=LOCALHOST, port=DPF_DEFAULT_PORT, timeout=10):
    """Launch ANSYS DPF.

    Parameters
    ----------
    ansys_path : str
        Full path to ANSYS.  For example:
        'C:\\Program Files\\ANSYS Inc\\v211'

    ip : str, optional
        IP address of the server.  Default to localhost.

    port : int
        Port of the server.  Defaults to the default DPF port
        ``50054``.

    Returns
    -------
    process : subprocess.Popen
        DPF Process.
    """
    if os.name == 'nt':
        run_cmd = f'Ans.Dpf.Grpc.bat --address {ip} --port {port}'
        path_in_install = "aisol/bin/winx64"
    else:
        run_cmd = ['./Ans.Dpf.Grpc.sh',f'--address {ip}',f'--port {port}']
        path_in_install = "aisol/bin/linx64"

    # verify ansys path is valid
    if os.path.isdir(f'{ansys_path}/{path_in_install}'):
        dpf_run_dir = f'{ansys_path}/{path_in_install}'
    else:
        dpf_run_dir = f'{ansys_path}'
    if not os.path.isdir(dpf_run_dir):
        raise NotADirectoryError(f'Invalid ansys path at "{ansys_path}".  '
                                 'Unable to locate the directory containing DPF at '
                                 f'"{dpf_run_dir}"')

    old_dir = os.getcwd()
    os.chdir(dpf_run_dir)
    process = subprocess.Popen(run_cmd,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    os.chdir(old_dir)

    # check to see if the service started
    lines = []
    def read_stdout():
        for line in io.TextIOWrapper(process.stdout, encoding="utf-8"):
            LOG.debug(line)
            lines.append(line)

    errors = []
    def read_stderr():
        for line in io.TextIOWrapper(process.stderr, encoding="utf-8"):
            LOG.error(line)
            errors.append(line)


    # must be in the background since the process reader is blocking
    Thread(target=read_stdout, daemon=True).start()
    Thread(target=read_stderr, daemon=True).start()

    t_timeout = time.time() + timeout
    started = False
    while not started:
        started = any('server started' in line for line in lines)

        if time.time() > t_timeout:
            raise TimeoutError(f'Server did not start in {timeout} seconds')

    # verify there were no errors
    time.sleep(1)
    if errors:
        try:
            process.kill()
        except PermissionError:
            pass
        errstr = '\n'.join(errors)
        if 'Only one usage of each socket address' in errstr:
            raise errors.InvalidPortError(f'Port {port} in use')
        raise RuntimeError(errstr)
        
    



"""
fields_container_factory
========================

Contains functions to make easy fields container creation.
"""

from ansys.dpf import core
from ansys.dpf.core import FieldsContainer, TimeFreqSupport
from ansys.dpf.core import fields_factory
from ansys.dpf.core.common import natures, locations
from ansys.grpc.dpf import base_pb2
from ansys.dpf.core import errors as dpf_errors

def over_time_freq_fields_container(fields, time_freq_unit = None, server = None):
    """Helper function to create a specific ``FieldsContainer``.
    The returned fields_container will contain one field by time set and, if needed, set
    the time freq support with the correct unit. 

    Parameters
    ----------
    fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary of Field entities to add to the fields container
    
    time_freq_unit : str , optional
        String that defines the unit symbol of the time_freq_support. Will be taken 
        into account if the fields attribute has a dictionary type.
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    fields_container : FieldsContainer
        FieldsContainer containing one field by time step.
    
    Examples
    --------
    Create a fields container based on time labels from scratch
    
    >>> from ansys.dpf import core as dpf
    >>> field1 = dpf.Field()
    >>> field2 = dpf.Field()
    >>> from ansys.dpf.core import fields_container_factory
    >>> my_fc = fields_container_factory.over_time_freq_fields_container([ field1, field2 ])    
    
    """
    if not isinstance(fields, dict) and not isinstance(fields, list):
        raise dpf_errors.InvalidTypeError("dictionary/list", "fields") 
    fc = FieldsContainer(server = server)
    fc.labels = { "time" }
    i = 0
    # dict case
    if isinstance(fields, dict):
        time_freq = []
        for field_key in fields:
            fc.add_field({ "time" : i + 1 }, fields[field_key])
            time_freq.append(field_key)
            i += 1
        time_freq_field = fields_factory.create_scalar_field(len(fields), location=locations.time_freq, server = server)
        time_freq_field.append(time_freq, 1)
        time_freq_field.unit = time_freq_unit
        time_freq_support = TimeFreqSupport(server = server)
        time_freq_support.time_frequencies = time_freq_field
        fc.time_freq_support = time_freq_support
    # list case
    elif isinstance(fields, list):
        for field in fields:
            fc.add_field({ "time" : i + 1 }, field)
            i += 1
    return fc

def over_time_freq_complex_fields_container(real_fields, imaginary_fields, time_freq_unit = None, server = None):
    """Helper function to create a specific ``FieldsContainer``.
    The returned fields_container will contain two fields (real and imaginary 
    fields) by time set.
    It sets the time freq support with the correct unit if needed (if the fields inputs 
    are dictionaries).

    Parameters
    ----------
    real_fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary or list of Field entities to add to the fields container
        
    imaginary_fields : Dictionary(time_int_key : Field) or list of Field
        Dictionary or list of Field entities to add to the fields container
        
    time_freq_unit : str , optional
        String that defines the unit symbol of the time_freq_support. Will be taken 
        into account if the fields attribute has a dictionary type.
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    fields_container : FieldsContainer
        FieldsContainer containing two fields (real and imaginary ones)
        by time step.
    """
    if not isinstance(real_fields, dict) and not isinstance(real_fields, list):
        raise dpf_errors.InvalidTypeError("dictionary/list", "real_fields") 
    if not isinstance(imaginary_fields, dict) and not isinstance(imaginary_fields, list):
        raise dpf_errors.InvalidTypeError("dictionary/list", "imaginary_fields") 
        
    errorString = "Both real_fields and imaginary_fields must have the same type (list or dict)"
    if isinstance(real_fields, dict):
        if not isinstance(imaginary_fields, dict):
            raise dpf_errors.DpfValueError(errorString)
    elif isinstance(real_fields, list):
        if not isinstance(imaginary_fields, list):
            raise dpf_errors.DpfValueError(errorString)
            
    fc = FieldsContainer(server = server)
    fc.labels = ["time", "complex"]
    i = 0
    # dict case
    if isinstance(real_fields, dict):
        time_freq = []
        for field_key in real_fields:
            fc.add_field({"time": i + 1, "complex" : 0}, real_fields[field_key])
            time_freq.append(field_key)
            i += 1
        i = 0
        im_time_freq = []
        for field_key in imaginary_fields:
            fc.add_field({"time": i + 1, "complex" : 1}, imaginary_fields[field_key])
            im_time_freq.append(field_key)
            i += 1
        time_freq_field = fields_factory.create_scalar_field(len(real_fields), locations.time_freq, server = server)
        time_freq_field.append(time_freq, 1)
        time_freq_field.unit = time_freq_unit
        im_time_freq_field = fields_factory.create_scalar_field(len(imaginary_fields), locations.time_freq, server = server)
        im_time_freq_field.append(im_time_freq, 1)
        im_time_freq_field.unit = time_freq_unit
        time_freq_support = TimeFreqSupport(server = server)
        time_freq_support.time_frequencies = time_freq_field
        time_freq_support.complex_frequencies = im_time_freq_field
        fc.time_freq_support = time_freq_support
    # list case
    if isinstance(real_fields, list):
        for field in real_fields:
            fc.add_field({"time": i + 1, "complex" : 0}, field)
            i += 1
        i = 0
        for field in imaginary_fields:
            fc.add_field({"time": i + 1, "complex" : 1}, field)
            i += 1
    return fc
    

def complex_fields_container(real_field, imaginary_field, server = None):
    """Helper function to create a specific ``FieldsContainer``.
    The returned fields_container will contain two fields (real and imaginary 
    fields) and only one time set.

    Parameters
    ----------
    real_fields : Field
        Real ansys.dpf.core.Field entity to add to the fields container
        
    imaginary_fields : Field
        Imaginary ansys.dpf.core.Field entity to add to the fields container
    
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.   
        
    Returns
    -------
    fields_container : FieldsContainer
        FieldsContainer containing two fields (real and imaginary ones).
    """
    fc = FieldsContainer(server = server)
    fc.labels = ["complex"]
    fc.add_field({ "complex" : 0 }, real_field)
    fc.add_field({ "complex" : 1 }, imaginary_field)
    return fc


"""
.. _ref_results:
    
Results
========
Module contains the Results and Result class that are created by the Model
to easily access results in result files."""
import functools


from ansys.dpf.core import Operator
from ansys.dpf.core import errors
from ansys.dpf.core.scoping import Scoping
from ansys.dpf.core.custom_fields_container import ElShapeFieldsContainer, BodyFieldsContainer

class Results:
    """Organize the results from DPF into accessible methods. All the available
    results are dynamically created depending on the model's 'ResultInfo'
    
    Attributes
    ----------
    displacement : Result 
        Result provider helper wrapping regular displacement operator. 
        With this wrapper, time and mesh scopings can easily
        be customized
                
        Examples
        --------
        Create a displacement result from the model
        and choose its time and mesh scopings
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement.on_last_time_freq.on_named_selection('_CONSTRAINEDNODES')
        >>> last_time_disp = disp.eval()
        
    stress : Result
        Result provider helper wrapping regular stress operator. 
        With this wrapper, time and mesh scopings, location.. can easily
        be customized
        
        Examples
        --------
        Create a stress result from the model
        and choose its time and mesh scopings
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> stress = model.results.stress.on_last_time_freq.on_named_selection('_CONSTRAINEDNODES')
        >>> last_time_stress = stress.eval()

    .... all other results : Result
        Result provider helper wrapping all types of provider available for a 
        given result file
        
        Examples
        --------
        Create a stress result from the model
        and choose its time and mesh scopings
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.electric_therm)
        >>> v = model.results.electric_potential
        >>> rf = model.results.reaction_force
        >>> dissip = model.results.thermal_dissipation_energy
        
    Examples
    --------
    Extract the result object from a model.

    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.simple_bar)
    >>> results = model.results # printable object

    Access the displacement at all times

    >>> from ansys.dpf.core import Model
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = Model(transient)
    >>> displacements = model.results.displacement.on_all_time_freqs.eval()

    """
    
    def __init__(self, model):
        self.__class__ = type(Results.__name__ + str(id(self)),(Results,),{})
        
        self._result_info = model.metadata.result_info
        self._model = model
        self._connect_operators()
        
    
    def __result__(self,result_type,*args):
        return Result(self._model,result_type)  
              
    def _connect_operators(self):
        """Dynamically add operators for results.

        The new operators subresults are connected to the model's
        streams.

        Examples
        --------
        >>> from ansys.dpf.core import Model
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = Model(transient)
        >>> disp_operator = model.results.displacement()
        >>> stress_operator = model.results.stress()
        >>> disp_x = model.results.displacement().X() 
        >>> disp_y = model.results.displacement().Y() 
        >>> disp_z = model.results.displacement().Z()
        
        """
        if self._result_info is None:
            return
        # dynamically add function based on input type
        self._op_map_rev = {}
        for result_type in self._result_info:
            try:                   
                doc =  Operator(result_type.operator_name, server=self._model._server).__str__()
                bound_method = self.__result__
                method2 = functools.partial(bound_method,
                                            result_type)
                setattr(self.__class__, result_type.name,property(method2,  doc=doc))
                
                self._op_map_rev[result_type.name] = result_type.name
            except errors.DPFServerException:
                pass            
            except Exception as e:
                print(result_type.name)
                raise e
        
   
    def __str__(self):
        return str(self._result_info)
    
    def __iter__(self):
        for key in self._op_map_rev:
            yield self.__class__.__dict__[key].fget()
            
            
    def __getitem__(self, val):
        n=0
        for key in self._op_map_rev:
            if n==val:
                return self.__class__.__dict__[key].fget()
            n+=1
                
    def __len__(self):
        return len(self._op_map_rev)
                
                
class Result:
    """An instance of this class helps using dpf's result providers.
    It helps connecting common inputs to the operator and 
    recovering its fields container output.
    'Result' is created by the model.
    
    Examples
    --------
    Create a displacement result from the model
    and choose its time and mesh scopings
    
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.msup_transient)
    >>> disp = model.results.displacement.on_last_time_freq.on_named_selection('_CONSTRAINEDNODES')
    >>> last_time_disp = disp.eval()
    
    Create a stress result from the model
    and split the result by element shapes: solid, shell, beam
    
    >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
    >>> stress = model.results.stress
    >>> stress_split = stress.split_by_shape.eval()
    >>> solid_stress = stress_split.solid_field()
    
    Create a strain result from the model on all time sets
    and recover the operator to connect it to other operators
    
    >>> model = dpf.Model(examples.msup_transient)
    >>> strain = model.results.elastic_strain.on_all_time_freqs()
    >>> eqv = dpf.operators.invariant.von_mises_eqv_fc(strain)
    >>> strain_eqv = eqv.outputs.fields_container()
    
    """
    
    def __init__(self, model,result_info):
        self._model= model
        self._time_scoping = None
        self._mesh_scoping= None
        self._location = None
        self._result_info = result_info
        self._specific_fc_type = None
        from ansys.dpf.core import operators
        try:
            #create the operator to read its documentation
            #if the operator doesn't exist, the method will not be added                
            doc =  Operator(self._result_info.operator_name, server=self._model._server).__str__()
            self.__doc__=doc
            if hasattr(operators, "result") and  hasattr(operators.result, self._result_info.name) :
                self._operator= getattr(operators.result, self._result_info.name)(server = self._model._server)
            else :
                self._operator= Operator(self._result_info.operator_name,server = self._model._server)
            self._operator._add_sub_res_operators(self._result_info.sub_results)
            self._model.__connect_op__(self._operator)
        except errors.DPFServerException:
            pass            
        except Exception as e:
            print(self._result_info.name)
            raise e
    
    def __call__(self,time_scoping=None, mesh_scoping=None):
        op= self._operator
        if time_scoping:
            op.inputs.time_scoping(time_scoping)            
        elif self._time_scoping:
            op.inputs.time_scoping(self._time_scoping)
            
        if mesh_scoping:
            op.inputs.mesh_scoping(mesh_scoping)
        elif self._mesh_scoping:
            op.inputs.mesh_scoping(self._mesh_scoping)
            
        if self._location:
            op.inputs.requested_location(self._location)
            
        return op
    
    def eval(self):
        """Evaluate the result provider with the inputs specified before 
        and returns its result fields container
        
        Returns
        -------
        fields_container : FieldsContainer, ElShapeFieldsContainer, BodyFieldsContainer
            if split_by_body is used, a BodyFieldsContainer is returned.
            if split_by_shape is used, a ElShapeFieldsContainer is returned.
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> fc = disp.on_all_time_freqs.eval()
        
        """
        fc = self.__call__().outputs.fields_container()
        if self._specific_fc_type =="shape":
            fc = ElShapeFieldsContainer(fields_container=fc,server=fc._server)
        elif self._specific_fc_type =="body":
            fc = BodyFieldsContainer(fields_container=fc,server=fc._server)
        return fc

    @property
    def on_all_time_freqs(self):
        """Sets the time scoping to all the time frequencies available in the time freq support
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_all_time_freqs.eval().get_label_scoping("time").ids
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20]
        
        """
        self._time_scoping = list(range(1,len(self._model.metadata.time_freq_support.time_frequencies)+1))
        return self
    
    @property
    def on_first_time_freq(self):
        """Sets the time scoping to the first time frequency available in the time freq support
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_first_time_freq.eval().get_label_scoping("time").ids
        [1]
        
        """
        self._time_scoping = 1
        return self
    
    @property
    def on_last_time_freq(self):
        """Sets the time scoping to the last time frequency available in the time freq support
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)
        >>> disp = model.results.displacement
        >>> disp.on_last_time_freq.eval().get_label_scoping("time").ids
        [20]
        
        """
        self._time_scoping =len(self._model.metadata.time_freq_support.time_frequencies)
        return self
    
    def on_time_scoping(self, time_scoping):
        """Sets the time scoping to the given one
        
        Parameters
        ----------
        time_scoping :  float, list[float], int, list[int], Scoping
            time/frequency(ies) or time set(s) are specified
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        Choose time sets
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)        
        >>> stress = model.results.stress
        >>> fc = stress.on_time_scoping([1,2,3,19]).eval()
        >>> len(fc)
        4
        
        Choose times, if the times chosen are not in the time freq support,
        results will be extrapolated
        
        >>> fc = stress.on_time_scoping([0.115,0.125]).eval()
        >>> len(fc)
        2
        >>> fc.time_freq_support.time_frequencies.data
        array([0.115, 0.125])
        
        """
        self._time_scoping = time_scoping
        return self
        
    
    def on_named_selection(self, named_selection):
        """Sets the mesh scoping to a given name selection's name
        
        Parameters
        ----------
        named_selection : str
            name of the named selection or component in upper case
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        Add a requested location to average result on nodes
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.msup_transient)        
        >>> stress = model.results.stress
        >>> fc = stress.on_first_time_freq.on_named_selection('_CONSTRAINEDNODES').eval()
        >>> len(fc[0].scoping)
        40
        
        """
        
        self._mesh_scoping = self._model.metadata.named_selection(named_selection)
        return self
    
    
    @property
    def split_by_body(self):
        """Set the mesh scoping to a scopings container
        where each scoping is a body
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
        >>> disp = model.results.displacement
        >>> fc_disp = disp.split_by_body.eval()
        >>> len(fc_disp)
        11
        >>> fc_disp.get_mat_scoping().ids  
        [1, 5, 6, 10, 2, 7, 8, 13, 4, 12, 15]
        >>> disp_mat_10 = fc_disp.get_field_by_mat_id(10)
        
        """
        self._specific_fc_type = "body"
        return self._add_split_on_property_type("mat")
    
    
    @property
    def split_by_shape(self):
        """Set the mesh scoping to a scopings container
        where each scoping is a specific element shape.
        The evaluated fields container will have on field on 'solid',
        one on 'shell', one on 'beam' and one on 'unknown_shape'
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_all_kinds_of_complexity())
        >>> disp = model.results.displacement
        >>> fc_disp = disp.split_by_shape.eval()
        >>> len(fc_disp)
        4
        
        >>> shell_disp = fc_disp.shell_field()
        >>> solid_disp = fc_disp.solid_field()
        
        """
        self._specific_fc_type = "shape"
        return self._add_split_on_property_type("elshape")
    
    def _add_split_on_property_type(self,prop):
        previous_mesh_scoping = self._mesh_scoping
        from ansys.dpf.core import operators
        if hasattr(operators, "scoping") and  hasattr(operators.scoping, "split_on_property_type") :
            self._mesh_scoping = operators.scoping.split_on_property_type()
        else: 
            self._mesh_scoping = Operator("scoping::by_property")
            
        self._mesh_scoping.inputs.requested_location(self._result_info.native_scoping_location)
        self._mesh_scoping.inputs.mesh(self._model.metadata.mesh_provider)
        self._mesh_scoping.inputs.label1(prop)
        if previous_mesh_scoping:
            try:
                self._mesh_scoping.inputs.mesh_scoping(previous_mesh_scoping)
            except :
                pass
        return self
    
    
    def on_mesh_scoping(self, mesh_scoping):
        """Set the mesh scoping to the given mesh_scoping
        
        Parameters
        ----------
        mesh_scoping : Scoping, list[int]
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        Use a list of nodes
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.complex_rst)
        >>> disp = model.results.displacement
        >>> fc = disp.on_mesh_scoping([1,2,3]).eval()
        >>> len(fc[0].scoping)
        3
        
        Use a Scoping to specify a list of entity ids with their location
        
        >>> stress = model.results.stress
        >>> scop = dpf.Scoping(ids=[3,4,5], location= dpf.locations.nodal)
        >>> fc = stress.on_mesh_scoping(scop).eval()
        >>> len(fc[0].scoping)
        3
        >>> fc[0].location
        'Nodal'

        """
        if isinstance(mesh_scoping, list):
            mesh_scoping= Scoping(ids=mesh_scoping,location=self._result_info.native_scoping_location,
                                  server=self._model._server)
            
        self._mesh_scoping = mesh_scoping
        return self
    
    
    def on_location(self, location):
        """Set the requested location of the provider:
        elemental nodal fields can be averaged to nodal or elemental location 
        
        Parameters
        ----------
        location : str, locations
        
        Returns
        -------
        self : Result
        
        Examples
        --------
        Add a requested location to average result on nodes
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.complex_rst)        
        >>> stress = model.results.stress
        >>> fc = stress.eval()
        >>> fc[0].location
        'ElementalNodal'

        >>> fc = stress.on_location(dpf.locations.nodal).eval()
        >>> fc[0].location
        'Nodal'

        """
        self._location = location
        return self


"""
FieldDefinition
===============
"""

from ansys import dpf
from ansys.grpc.dpf import (base_pb2,
                            field_definition_pb2, field_definition_pb2_grpc)
from ansys.dpf.core.common import natures, shell_layers
from ansys.dpf.core.dimensionnality import Dimensionnality


class FieldDefinition:
    """Contains the physical and mathematical description of the Field:
    unit, homogeneity, dimensionnnality...                        
    """

    def __init__(self, field_definition=None, server=None):          
        if server is None:
            server = dpf.core._global_server()

        self._server = server
        self._stub = self._connect(self._server.channel)
        if isinstance(field_definition, field_definition_pb2.FieldDefinition):
            self._messageDefinition = field_definition
        else:
            request = base_pb2.Empty()
            self._messageDefinition = self._stub.Create(request)
            
    @property
    def location(self):
        """Location
        
        Returns
        -------
        location : str
            where the data is located (Nodal, Elemental, TimeFreq_sets...)
        """
        out = self._stub.List(self._messageDefinition)
        return out.location.location

    @property
    def unit(self):
        """Returns the unit of the field

        Returns
        -------
        unit : str
        """
        return self._stub.List(self._messageDefinition).unit.symbol
    

    @property
    def shell_layers(self):        
        """Shell_layers
        
        Returns
        -------
        shell_layers : shell_layers
            returns LayerIndependent for fields unrelated to layers
        """
        enum_val = self._stub.List(self._messageDefinition).shell_layers
        return shell_layers(enum_val.real-1) #+1 is added to the proto enum to have notset as 0
    
    @property
    def dimensionnality(self):
        """Dimensionnality
        
        Returns
        -------
        dimensionnality : Dimensionnality
            nature and size of the elementary data
        """
        val = self._stub.List(self._messageDefinition).dimensionnality
        return Dimensionnality(val.size, natures(val.nature.real))
    
    @unit.setter
    def unit(self, value):
        self._modify_field_def(unit=value)
        
        
    @location.setter
    def location(self, value):
        self._modify_field_def(location=value)
        
        
    @shell_layers.setter
    def shell_layers(self, value):
        self._modify_field_def(shell_layer=value)
        
        
    @dimensionnality.setter
    def dimensionnality(self, value):
        self._modify_field_def(dimensionnality=value)
    
    
    def _modify_field_def(self, unit = None, location = None, dimensionnality = None, shell_layer=None):
        request = field_definition_pb2.FieldDefinitionUpdateRequest()
        request.field_definition.CopyFrom(self._messageDefinition)
        if unit != None:
            request.unit_symbol.symbol = unit
        if location != None:
            request.location.location = location
        if dimensionnality != None:
            if not isinstance(dimensionnality, Dimensionnality):
                raise TypeError("the dimensionnality needs to be of type Dimensionnsality")
            request.dimensionnality.CopyFrom(dimensionnality._parse_dim_to_message())
        if shell_layer != None:
            if isinstance(shell_layer, shell_layers):
                request.shell_layers = shell_layer.value+1
            else:                
                request.shell_layers = shell_layer+1
        self._stub.Update(request)
        
    
    def __del__(self):
        try:
            self._stub.Delete(self._messageDefinition)
        except:
            pass

    def _connect(self, channel):
        """Connect to the grpc service"""
        return field_definition_pb2_grpc.FieldDefinitionServiceStub(channel)



"""Miscellaneous functions for DPF module"""
import platform
import glob
import os
from pkgutil import iter_modules


DEFAULT_FILE_CHUNK_SIZE =65536

# ANSYS CPython workbench environment may not have scooby installed
try:
    from scooby import Report as ScoobyReport
except ImportError:
    class ScoobyReport():
        """Placeholder for scooby.Report"""

        def __init__(self, *args, **kwargs):
            raise ImportError('Install `scooby` with `pip install scooby` to use '
                              'this feature')


def module_exists(module_name):
    """Returns True when a module exists"""
    return module_name in (name for loader, name, ispkg in iter_modules())


class Report(ScoobyReport):
    """Generate a report of the installed packages for ansys-dpf-core"""

    def __init__(self, additional=None, ncol=3, text_width=80, sort=False,
                 gpu=True):
        """Generate a :class:`scooby.Report` instance.

        Parameters
        ----------
        additional : list(ModuleType), list(str)
            List of packages or package names to add to output information.

        ncol : int, optional
            Number of package-columns in html table; only has effect if
            ``mode='HTML'`` or ``mode='html'``. Defaults to 3.

        text_width : int, optional
            The text width for non-HTML display modes

        sort : bool, optional
            Alphabetically sort the packages

        gpu : bool
            Gather information about the GPU. Defaults to ``True`` but if
            experiencing renderinng issues, pass ``False`` to safely generate
            a report.

        """

        # Mandatory packages.
        core = ['ansys.grpc.dpf']

        # Optional packages.
        optional = []

        # Information about the GPU - bare except in case there is a rendering
        # bug that the user is trying to report.
        if gpu:
            from pyvista.utilities.errors import GPUInfo
            try:
                extra_meta = [(t[1], t[0]) for t in GPUInfo().get_info()]
            except:
                extra_meta = ("GPU Details", "error")
        else:
            extra_meta = ("GPU Details", "None")

        super().__init__(additional=additional, core=core,
                         optional=optional, ncol=ncol,
                         text_width=text_width, sort=sort,
                         extra_meta=extra_meta)


def is_float(string):
    """Returns true when a string can be converted to a float"""
    try:
        float(string)
        return True
    except ValueError:
        return False


def is_ubuntu():
    """True when running Ubuntu"""
    if os.name == 'posix':
        return 'ubuntu' in platform.platform().lower()
    return False


def find_ansys():
    """Searches for ansys path within the standard install location
    and returns the path of the latest version.
    
    Returns
    -------
    ansys_path : str
        Full path to ANSYS.  For example, on windows:

    Examples
    --------
    Within Windows

    >>> from ansys.dpf.core.misc import find_ansys
    >>> path = find_ansys()
    
    Within Linux

    >>> path = find_ansys()
    
    """
    base_path = None
    if os.name == 'nt':
        base_path = os.path.join(os.environ['PROGRAMFILES'], 'ANSYS INC')
    elif os.name == 'posix':
        for path in ['/usr/ansys_inc', '/ansys_inc']:
            if os.path.isdir(path):
                base_path = path
    else:
        raise OSError(f'Unsupported OS {os.name}')

    if base_path is None:
        return base_path

    paths = glob.glob(os.path.join(base_path, 'v*'))

    if not paths:
        return None

    versions = {}
    for path in paths:
        ver_str = path[-3:]
        if is_float(ver_str):
            versions[int(ver_str)] = path

    return versions[max(versions.keys())]



"""
.. _ref_field:
    
Field
=====
"""
import numpy as np

from ansys import dpf
from ansys.grpc.dpf import (field_pb2, field_pb2_grpc, base_pb2)
from ansys.dpf.core.common import natures, types, locations
from ansys.dpf.core import scoping, meshed_region, time_freq_support
from ansys.dpf.core.plotter import Plotter
from ansys.dpf.core import errors
from ansys.dpf.core.field_definition import FieldDefinition
from ansys.dpf.core.field_base import _FieldBase, _LocalFieldBase
from ansys.dpf.core.dimensionnality import Dimensionnality

class Field(_FieldBase):
    """Class representing the main simulation data container.
    It can be evaluated data from a ``ansys.dpf.core.Operator``
    or created (by a factory and directly by an instance of this class).
    Field's data is always associated to its scoping (entities associated to each value)
    and support (subset of the model where is the data), 
    making the field a self-describing piece of data.

    Parameters
    ----------
    nentities : int
        Number of entities reserved

    nature : natures, optional
        Nature of the field.

    location : str, optional
        Location of the field.  For example:

        - ``"Nodal"``
        - ``"Elemental"``
        - ``"ElementalNodal"``

    field : ansys.grpc.dpf.field_pb2.Field, optional
        Field message generated from a grpc stub.
        
    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    >>> # 1. Create a field from scratch
    >>> from ansys.dpf.core import fields_factory
    >>> from ansys.dpf.core import locations
    >>> from ansys.dpf import core as dpf
    >>> field_with_classic_api = dpf.Field()
    >>> field_with_classic_api.location = locations.nodal
    >>> field_with_factory = fields_factory.create_scalar_field(10)
    
    >>> # 2. Extract a displacement field from a transient result file.
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = dpf.Model(transient)
    >>> disp = model.results.displacement()
    >>> fields_container = disp.outputs.fields_container()
    >>> field = fields_container[0]
    >>> len(field)
    11460
    >>> field.component_count
    3
    >>> field.elementary_data_count
    3820
        
    >>> # 3. Create a displacement field
    >>> from ansys.dpf import core as dpf
    >>> import numpy as np
    >>> my_field = dpf.Field(10, dpf.natures.vector,dpf.locations.nodal)
    >>> my_field.data = np.zeros(30)
    >>> my_field.scoping.ids = range(1,11)
    
    >>> # 4. Set data
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = dpf.Model(transient)
    >>> disp = model.results.displacement()
    >>> fields_container = disp.outputs.fields_container()
    >>> field = fields_container[0]
    >>> field.data[2]
    array([-0.00672665, -0.03213735,  0.00016716])
    
    """
    
    def __init__(self, nentities=0, nature=natures.vector,
                 location=locations.nodal, field=None, server=None):
        """Initialize the field with either optional field message, or
        by connecting to a stub.
        """
        super().__init__(nentities, nature, location, False, field, server)
        self._field_definition = self._load_field_definition() 
        
        
    def as_local_field(self):
        """Creates a deep copy of the field locally so that the user can access 
        and modify it locally without any request sent to the server.
        This should be used in a with statement, so that the local field
        is released and the data is sent to the server in one shot.
        If it's not used in a with statement, the method release_data()
        should be used to actually update the field.
        
        Warning
        -------
        If this as_local_field method is not used as a context manager in a 
        with statement or if the method release_data() is not called,
        the data will not be actually updated.
        
        Returns
        -------
        local_field : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> num_entities = 3
        >>> field_to_local = dpf.fields_factory.create_3d_vector_field(num_entities, location=dpf.locations.elemental_nodal)
        >>> with field_to_local.as_local_field() as f:    
        ...     for i in range(1,num_entities+1):
        ...         f.append([[0.1*i,0.2*i, 0.3*i],[0.1*i,0.2*i, 0.3*i]],i)
        ...         f.get_entity_data(i-1),[[0.1*i,0.2*i, 0.3*i],[0.1*i,0.2*i, 0.3*i]]
        (array([[0.1, 0.2, 0.3],
               [0.1, 0.2, 0.3]]), [[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]])
        (array([[0.2, 0.4, 0.6],
               [0.2, 0.4, 0.6]]), [[0.2, 0.4, 0.6], [0.2, 0.4, 0.6]])
        (array([[0.3, 0.6, 0.9],
               [0.3, 0.6, 0.9]]), [[0.30000000000000004, 0.6000000000000001, 0.8999999999999999], [0.30000000000000004, 0.6000000000000001, 0.8999999999999999]])
            
        """
        return _LocalField(self)  
    
   
    @property
    def location(self):
        """Return the field location.

        Returns
        -------
        location : str
            Location string.  Either ``'Nodal'``, ``'Elemental'``, or
            ``'ElementalNodal'``.

        Examples
        --------
        Location for a stress field evaluated at nodes

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_transient_result())
        >>> s_op = model.results.stress()
        >>> s_fc = s_op.outputs.fields_container()
        >>> field = s_fc[0]
        >>> field.location
        'ElementalNodal'
        
        """
        if self.field_definition:
            return self.field_definition.location
        
    @location.setter
    def location(self, value):
        """Change the field location.

        Parameters
        -------
        location : str or locations
            Location string.  Either ``'Nodal'``, ``'Elemental'``, 
            ``'ElementalNodal'``...

        Examples
        --------
        Location for a field evaluated at nodes

        >>> from ansys.dpf import core as dpf
        >>> import numpy as np
        >>> my_field = dpf.Field(10, dpf.natures.vector,dpf.locations.nodal)
        >>> my_field.data = np.zeros(30)
        >>> my_field.scoping.ids = range(1,11)
        >>> my_field.location
        'Nodal'
        >>> my_field.location = dpf.locations.elemental_nodal
        >>> my_field.location
        'ElementalNodal'
        
        """
        fielddef = self.field_definition
        fielddef.location = value
        self.field_definition = fielddef

    @property
    def shell_layers(self):
        """Return the field's shell layers order (or lack of shell layers) 
        that defines how the field's data is ordered.

        Returns
        -------
        Enum : shell_layers
        """
        if self.field_definition:
            return self.field_definition.shell_layers
        
    @shell_layers.setter
    def shell_layers(self, value):
        """Change the field's shell layers.

        Parameters
        -------
        Enum : shell_layer
        """
        fielddef = self.field_definition
        fielddef.shell_layers = value
        self.field_definition = fielddef

    def to_nodal(self):
        """Convert this field to one with a Nodal location.

        Only valid when this field's location is ElementalNodal or
        Elemental.

        Returns
        -------
        nodal_field : Field 
            with ``location=='Nodal'``.
        """
        if self.location == 'Nodal':
            raise errors.LocationError('Location is already "Nodal"')

        op = dpf.core.Operator("to_nodal")
        op.inputs.connect(self)
        return op.outputs.field()

    def plot(self, notebook=None, shell_layers=None):
        """Plot the field/fields container on mesh support if exists.

        Warning
        -------
        This is primarily added out of convenience as plotting
        directly from the field can be slower than extracting the
        meshed region and plotting the field on top of that.  It is
        more efficient to plot with:

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> mesh = model.metadata.meshed_region
        >>> disp = model.results.displacement()
        >>> fields_container = disp.outputs.fields_container()
        >>> field = fields_container[0]
        >>> mesh.plot(field)
        [(1.675707321585373, 1.675707321585373, 2.425707321585373),
         (0.0, 0.0, 0.75),
         (0.0, 0.0, 1.0)]

        Parameters
        ----------
        notebook : bool, optional
            Bool, that specifies if the plotting is in the notebook as
            a static image or or as a dynamic plot outside of the
            notebook.

        shell_layers : shell_layers, optional
            Enum used to set the shell layers if the model to plot
            contains shell elements.
        """
        pl = Plotter(self.meshed_region)
        pl.plot_contour(self, notebook, shell_layers)

    def resize(self, nentities, datasize):
        """Allocate memory.

        Parameters
        ----------
        nentities : int
            num ids in the scoping

        datasize : int
            data vector size
        """
        request = field_pb2.UpdateSizeRequest()
        request.field.CopyFrom(self._message)
        request.size.scoping_size = nentities
        request.size.data_size = datasize
        self._stub.UpdateSize(request)

    def _load_field_definition(self):
        """Attempt to load the field definition for this field"""
        try:
            request=field_pb2.GetRequest()
            request.field.CopyFrom(self._message)
            out = self._stub.GetFieldDefinition(request)
            return FieldDefinition(out.field_definition, self._server)
        except:
            return

    @property
    def unit(self):
        """Units of the field

        Returns
        ----------
        unit : str
        
        Examples
        --------
        Units of a displacement field

        >>> from ansys.dpf import core as dpf
        >>> my_field = dpf.Field(10, dpf.natures.vector,dpf.locations.nodal)
        >>> my_field.unit = "m"
        >>> my_field.unit
        'm'
        
        """
        if self.field_definition:
            return self.field_definition.unit
        
    @unit.setter
    def unit(self, value):
        """Change the unit of the field
        
        Parameters
        ----------
        unit : str

        Examples
        --------
        Units of a displacement field

        >>> from ansys.dpf import core as dpf
        >>> my_field = dpf.Field(10, dpf.natures.vector,dpf.locations.nodal)
        >>> my_field.unit = "m"
        >>> my_field.unit
        'm'
        
        """
        fielddef = self.field_definition
        fielddef.unit = value
        self.field_definition = fielddef
        
    @property
    def dimensionnality(self):
        """The Dimensionnality represents the shape of an elementary
        data contains in the Field.
        
        Returns
        -------
        dimensionnality : Dimensionnality
            nature and size of the elementary data
        """
        if self.field_definition:
            return self.field_definition.dimensionnality
        
    @dimensionnality.setter
    def dimensionnality(self, value):
        """
        Parameters
        ----------
        dimensionnality : dpf.core.Dimensionnality
            nature and size of the elementary data
        """
        fielddef = self.field_definition
        fielddef.dimensionnality = value
        self.field_definition = fielddef
        
    @property
    def name(self):
        request = field_pb2.GetRequest()
        request.field.CopyFrom(self._message)
        out = self._stub.GetFieldDefinition(request)
        return out.name
        
            
    def _set_field_definition(self, field_definition):
        """
        Parameters
        ----------
        field_definition : FieldDefinition
        """
        request = field_pb2.UpdateFieldDefinitionRequest()
        request.field_def.CopyFrom(field_definition._messageDefinition)
        request.field.CopyFrom(self._message)
        self._stub.UpdateFieldDefinition(request)
           
    @property
    def field_definition(self):
        """The field definition defines what is the field: its location, unit, dimensionnality, shell layers...
        """
        return self._field_definition
    
    
    @field_definition.setter
    def field_definition(self, value):
        """The field definition defines what is the field: its location, unit, dimensionnality, shell layers...

        Parameters
        ----------
        field_definition : FieldDefinition
        """
        return self._set_field_definition(value)
    
    def _get_meshed_region(self):
        """
        Returns
        -------
        meshed_region : MeshedRegion
        """
        request = field_pb2.SupportRequest()
        request.field.CopyFrom(self._message)
        request.type = base_pb2.Type.Value("MESHED_REGION")
        try:
            message = self._stub.GetSupport(request)
            return meshed_region.MeshedRegion(mesh=message, server = self._server)
        except:
            raise RuntimeError("The field's support is not a mesh.  Try a time_freq_support.")

    def _get_time_freq_support(self):
        """
        Returns
        -------
        time_freq_support : TimeFreqSupport
        """
        request = field_pb2.SupportRequest()
        request.field.CopyFrom(self._message)
        request.type = base_pb2.Type.Value("TIME_FREQ_SUPPORT")
        try:
            message = self._stub.GetSupport(request)
            return time_freq_support.TimeFreqSupport(time_freq_support=message, server = self._server)
        except:
            raise RuntimeError("The field's support is not a timefreqsupport.  Try a mesh.")
            
    def _set_support(self, support, support_type: str):
        """
        Set field support as time_freq_support or meshed_region.
        """
        request = field_pb2.SetSupportRequest()
        request.field.CopyFrom(self._message)
        request.support.type = base_pb2.Type.Value(support_type)
        request.support.id = support._message.id
        self._stub.SetSupport(request)

    @property
    def time_freq_support(self):
        """
        Time_freq_support of the associated field.
        
        Returns
        -------
        time_freq_support : TimeFreqSupport
            Describe the time frequency support of the field. 
        """
        return self._get_time_freq_support()
    
    @time_freq_support.setter
    def time_freq_support(self, value):
        """
        Set the time_freq_support of the associated field.
        
        Parameters
        ----------
        time_freq_support : TimeFreqSupport
            Describe the time frequency support of the field. 
        """
        self._set_support(value, "TIME_FREQ_SUPPORT")
        
    @property
    def meshed_region(self):
        """
        Meshed_region of the associated field.
        
        Return
        ------
        meshed_region : MeshedRegion
            Describe the mesh support of the field. 
        """
        return self._get_meshed_region()
    
    @meshed_region.setter
    def meshed_region(self, value):
        """
        Set the meshed_region of the associated field.
        
        Parameters
        ----------
        meshed_region : MeshedRegion
            Describe the mesh support of the field. 
        """
        self._set_support(value, "MESHED_REGION")

    def __add__(self, field_b):
        """Adds two fields together
                
        Returns
        -------
        add : operators.math.add
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "add") :
            op= operators.math.add()
        else :
            op= dpf_operator.Operator("add")
        op.connect(0,self)        
        op.connect(1, field_b)
        return op

    def __pow__(self, value):
        if value != 2:
            raise ValueError('DPF only the value is "2" suppported')
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "sqr") :
            op= operators.math.sqr(self)
        else :
            op= dpf_operator.Operator("sqr")
            op.connect(0,self)        
        return op
    
    def __mul__(self, value):
        """Multiplies two fields together
        
        Returns
        -------
        mul : operators.math.generalized_inner_product
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "generalized_inner_product") :
            op= operators.math.generalized_inner_product()
        else :
            op= dpf_operator.Operator("generalized_inner_product")
        op.connect(0,self)        
        op.connect(1, value)
        return op
    
    def __sub__(self, fields_b):
        """Subtract two fields together
                
        Returns
        -------
        minus : operators.math.minus
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "minus") :
            op= operators.math.minus()
        else :
            op= dpf_operator.Operator("minus")
        op.connect(0,self)        
        op.connect(1, fields_b)
        return op
    
    def _min_max(self):
        from ansys.dpf.core import dpf_operator
        op = dpf_operator.Operator("min_max")
        op.connect(0, self)
        return op

    def min(self):
        """Component-wise minimum over this field

        Returns
        -------
        min : Field
            Component-wise minimum field.
        """
        return self._min_max().get_output(0, types.field)

    def max(self):
        """Component-wise maximum over this field

        Returns
        -------
        max : Field
            Component-wise maximum field.
        """
        return self._min_max().get_output(1, types.field)
    
    def deep_copy(self,server=None):
        """Creates a deep copy of the field's data on a given server.
        This can be useful to pass data from one server instance to another.
        
        Parameters
        ----------
        server : DPFServer, optional
        
        Returns
        -------
        field_copy : Field
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> fields_container = disp.outputs.fields_container()
        >>> field = fields_container[0]
        >>> other_server = dpf.start_local_server(as_global=False)
        >>> deep_copy = field.deep_copy(server=other_server)
        
        """            
        f = Field(nentities=len(self.scoping), location=self.location,nature=self.field_definition.dimensionnality.nature, server=server)
        f.scoping = self.scoping.deep_copy(server)
        f.data = self.data
        f.unit = self.unit
        f.location = self.location
       
        try:
            f._data_pointer = self._data_pointer
        except:
            pass
        try:
            f.meshed_region = self.meshed_region.deep_copy(server=server)
        except:
            pass
        try:
            f.time_freq_support = self.time_freq_support.deep_copy(server=server)
        except:
            pass
        
        return f
        
class _LocalField(_LocalFieldBase,Field):
    """Class only created by a field to cache the internal data of the field,
    modify it locallly, and send a single update request to the server 
    when the local field is deleted
    
    Parameters
    ----------
    field : Field
        field to copy
        
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> import numpy as np
    >>> num_entities = 3
    >>> field_to_local = dpf.fields_factory.create_3d_vector_field(num_entities, location=dpf.locations.elemental_nodal)
    >>> with field_to_local.as_local_field() as f:    
    ...     for i in range(1,num_entities+1):
    ...         f.append(np.array([[0.1*i,0.2*i, 0.3*i],[0.1*i,0.2*i, 0.3*i]]),i)
    ...         f.get_entity_data(i-1),[[0.1*i,0.2*i, 0.3*i],[0.1*i,0.2*i, 0.3*i]]
    (array([[0.1, 0.2, 0.3],
           [0.1, 0.2, 0.3]]), [[0.1, 0.2, 0.3], [0.1, 0.2, 0.3]])
    (array([[0.2, 0.4, 0.6],
           [0.2, 0.4, 0.6]]), [[0.2, 0.4, 0.6], [0.2, 0.4, 0.6]])
    (array([[0.3, 0.6, 0.9],
           [0.3, 0.6, 0.9]]), [[0.30000000000000004, 0.6000000000000001, 0.8999999999999999], [0.30000000000000004, 0.6000000000000001, 0.8999999999999999]])

    """
    
    def __init__(self, field):
        super().__init__(field)
        


"""
AvailableResult
===============
"""

from warnings import warn
from ansys.grpc.dpf import available_result_pb2, base_pb2
from ansys.dpf.core.common import _remove_spaces

class AvailableResult:
    """A class used to represent an Available result which can be
    requested via an operator

    Parameters
    ----------
    availableresult : available_result_pb2.AvailableResult message
    
    Examples
    --------
    Explore an available result from the model
    
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> transient = examples.download_transient_result()
    >>> model = dpf.Model(transient)
    >>> result_info = model.metadata.result_info
    >>> res = result_info.available_results[0]
    >>> res.name
    'displacement'
    >>> res.homogeneity
    'length'
    >>> res.dimensionality
    'vector'
    
    Create the operator of the given available result    
    >>> disp = model.results.displacement()

    """

    def __init__(self, availableresult):
        """Initialize the AvailableResult with an availableResult message"""
        self._message = availableresult

    def __str__(self):
        txt = self.name+'\n' +\
              'Operator name: "%s"\n' % self.operator_name +\
              'Number of components: %d\n' % self.n_components +\
              'Dimensionality: %s\n' % self.dimensionality +\
              'Homogeneity: %s\n' % self.homogeneity
        if self.unit:
            txt += 'Units: %s\n' % self.unit
        return txt

    @property
    def name(self):
        """Result operator"""
        if self.operator_name in _result_properties:
            return _result_properties[self.operator_name]["scripting_name"]
        else:
            return _remove_spaces(self._message.physicsname)

    @property
    def n_components(self):
        """Number of components of the results"""
        return self._message.ncomp

    @property
    def dimensionality(self):
        """Dimensionality nature of the result (vector, scalar, tensor...)"""
        return base_pb2.Nature.Name(self._message.dimensionality).lower()

    @property
    def homogeneity(self):
        """Homogeneity of the result"""
        try:
            homogeneity = self._message.homogeneity
            if (homogeneity==117):
                return available_result_pb2.Homogeneity.Name(available_result_pb2.Homogeneity.DIMENSIONLESS).lower()
            return available_result_pb2.Homogeneity.Name(homogeneity).lower()
        except ValueError as exception:
            warn(str(exception))
            return ''

    @property
    def unit(self):
        """Unit of the result"""
        return self._message.unit.lower()

    @property
    def operator_name(self):
        """Name of the corresponding operator"
        """
        return self._message.name
    
    @property
    def sub_results(self):
        """List of sub result"
        """
        rep_sub_res =self._message.sub_res
        list = []
        for sub_res in rep_sub_res:
            try :
                int(sub_res.name)
                dict ={"name":"principal"+sub_res.name, "operator name":sub_res.op_name, "description":sub_res.description }
            except :
                dict ={"name":sub_res.name, "operator name":sub_res.op_name, "description":sub_res.description }
            list.append(dict)
        return list
    
    
    @property
    def native_location(self):
        if self.operator_name in _result_properties:
            return _result_properties[self.operator_name]["location"]
        
    @property
    def native_scoping_location(self):
        loc = self.native_location
        if loc == "ElementalNodal":
            return "Elemental"
        else:
            return loc
    
_result_properties ={ "S":{"location":"ElementalNodal", "scripting_name":"stress"},
            "ENF":{"location":"ElementalNodal", "scripting_name":"element_nodal_forces"},
            "EPEL":{"location":"ElementalNodal", "scripting_name":"elastic_strain"},
            "EPPL":{"location":"ElementalNodal", "scripting_name":"plastic_strain"},
            "ECR":{"location":"ElementalNodal", "scripting_name":"creep_strain"},
            "BFE":{"location":"ElementalNodal", "scripting_name":"structural_temperature"},
            "ETH":{"location":"ElementalNodal", "scripting_name":"thermal_strain"},
            "ETH_SWL":{"location":"ElementalNodal", "scripting_name":"swelling_strains"},
            "ENG_VOL":{"location":"Elemental", "scripting_name":"elemental_volume"},
            "ENG_SE":{"location":"Elemental", "scripting_name":"stiffness_matrix_energy"},
            "ENG_AHO":{"location":"Elemental", "scripting_name":"artificial_hourglass_energy"},
            "ENG_KE":{"location":"Elemental", "scripting_name":"kinetic_energy"},
            "ENG_CO":{"location":"Elemental", "scripting_name":"co_energy"},
            "ENG_INC":{"location":"Elemental", "scripting_name":"incremental_energy"},
            "ENG_TH":{"location":"Elemental", "scripting_name":"thermal_dissipation_energy"},
            "U":{"location":"Nodal", "scripting_name":"displacement"},
            "V":{"location":"Nodal", "scripting_name":"velocity"},
            "A":{"location":"Nodal", "scripting_name":"acceleration"},
            "RF":{"location":"Nodal", "scripting_name":"reaction_force"},
            "F":{"location":"Nodal", "scripting_name":"nodal_force"},
            "M":{"location":"Nodal", "scripting_name":"nodal_moment"},
            "TEMP":{"location":"Nodal", "scripting_name":"temperature"},
            "EF":{"location":"ElementalNodal", "scripting_name":"electric_field"},
            "VOLT":{"location":"Nodal", "scripting_name":"electric_potential"},
            "TF":{"location":"ElementalNodal", "scripting_name":"heat_flux"},
            "UTOT":{"location":"Nodal", "scripting_name":"raw_displacement"},
            "RFTOT":{"location":"Nodal", "scripting_name":"raw_reaction_force"}
            }




from ansys.grpc.dpf import field_pb2, base_pb2, field_pb2_grpc
from ansys.dpf.core import scoping
from ansys.dpf.core.common import natures, locations
from ansys.dpf.core import errors 
from ansys.dpf.core import server as serverlib

import numpy as np

class _FieldBase:
    """Base APIs for all implementations that follow Dpf's
    field concept."""
    
    
    def __init__(self, nentities=0, nature=natures.vector,
                 location=locations.nodal, is_property_field = False, 
                 field=None, server=None):
        """Initialize the field with either optional field message, or
        by connecting to a stub.
        """
        if server is None:
            server = serverlib._global_server()

        self._server = server
        self._stub = self._connect()

        if field is None:
            request = field_pb2.FieldRequest()
            if hasattr(nature, 'name'):
                snature = nature.name
            else:
                snature = nature
            request.nature = base_pb2.Nature.Value(snature.upper())
            request.location.location = location
            request.size.scoping_size = nentities
            if snature==natures.vector.name:
                elem_data_size =3
            elif snature==natures.symmatrix.name:
                elem_data_size =6
            else:
                elem_data_size=1
            request.size.data_size = nentities*elem_data_size
            if is_property_field:
                request.datatype = u"int"
            self._message = self._stub.Create(request)
        else:
            from ansys.dpf.core import field as field_module
            from ansys.dpf.core import property_field
            if isinstance(field, field_module.Field):
                self._message = field._message
            elif isinstance(field, property_field.PropertyField):
                self._message = field._message
            elif isinstance(field, field_pb2.Field):
                self._message = field
            else:
                raise TypeError(f'Cannot create a field from a "{type(field)}" object')
    @property
    def shape(self):
        """Numpy-like shape of the field
        
        Examples
        --------
        Shape of a stress field
        
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.download_transient_result())
        >>> s_op =model.results.stress()
        >>> s_fc = s_op.outputs.fields_container()
        >>> field = s_fc[0]
        >>> field.shape
        (5720, 6)
        
        """
        if self.component_count != 1:
            return (self.elementary_data_count, self.component_count)
        return self.elementary_data_count       
                
    @property
    def component_count(self):
        """Number of components in an elementary data of the Field
        
        Returns
        -------
        ncomp : int
            Number of component of the each elementary data
        """
        request = field_pb2.CountRequest()
        request.entity = base_pb2.NUM_COMPONENT
        request.field.CopyFrom(self._message)
        return self._stub.Count(request).count
    
    @property
    def elementary_data_count(self):
        """Number of elementary data in the field"""
        request = field_pb2.CountRequest()
        request.entity = base_pb2.NUM_ELEMENTARY_DATA
        request.field.CopyFrom(self._message)
        return self._stub.Count(request).count
    
    @property
    def size(self):
        """The length of the data vector.
        Also equals to the number of elementary data times the number of components.
        
        Returns
        -------
        size : int
        """
        return self.elementary_data_count*self.component_count
    
    @property
    def elementary_data_shape(self):
        """Numpy-like shape of the field"""
        if self.component_count != 1:
            return (1, self.component_count)
        else:
            return self.component_count
        
    @property
    def ndim(self):
        return self.component_count

    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """        
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)

    def __len__(self):
        return self.size
    
    def _del_scoping(self, scope):
        scope.__del__()

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def _connect(self):
        """Connect to the grpc service"""
        return field_pb2_grpc.FieldServiceStub(self._server.channel)
    
    def _set_scoping(self, scoping):
        """
        Parameters
        ----------
        scoping : Scoping
        """
        request = field_pb2.UpdateScopingRequest()
        request.scoping.CopyFrom(scoping._message)
        request.field.CopyFrom(self._message)
        self._stub.UpdateScoping(request)
        
    def _get_scoping(self):
        """
        Returns
        -------
        scoping : Scoping
        """
        request = field_pb2.GetRequest()
        request.field.CopyFrom(self._message)
        message = self._stub.GetScoping(request)
        return scoping.Scoping(scoping=message.scoping, server = self._server)

    @property
    def scoping(self):
        """ The scoping allows to know where is the data.
        Each entity data is on a given scoping id.
        
        Returns
        -------
        scoping : Scoping
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> stress_op = model.results.stress()
        >>> fields_container = stress_op.outputs.fields_container()
        >>> scoping = fields_container[0].scoping
        >>> scoping.location
        'Elemental'
        >>> scoping.id(3)
        586
        >>> #The fourth elementary data of the field corresponds to 
        >>> #the element id number 586 in the mesh
        """
        return self._get_scoping()
    
    @scoping.setter
    def scoping(self, scoping):
        """
        Parameters
        ----------
        scoping : Scoping
        """
        return self._set_scoping(scoping)
               
        
    def get_entity_data(self, index):
        """Returns the elementary data of the scoping's index in parameter
        
        Returns
        --------
        data : numpy.array
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> stress_op = model.results.stress()
        >>> fields_container = stress_op.outputs.fields_container()
        >>> fields_container[0].get_entity_data(0)
        array([[-3.27795062e+05,  1.36012200e+06,  1.49090608e+08,
                -4.88688900e+06,  1.43038560e+07,  1.65455040e+07],
               [-4.63817550e+06,  1.29312225e+06,  1.20411832e+08,
                -6.06617800e+06,  2.34829700e+07,  1.77231120e+07],
               [-2.35684860e+07, -3.53474400e+07,  2.01501168e+08,
                -5.23361700e+06, -2.88789280e+07, -6.16478200e+06],
               [-3.92756960e+07, -2.72369280e+07,  1.81454016e+08,
                -3.75441450e+06, -3.62480300e+06, -3.26075620e+07],
               [ 1.63554530e+07,  2.83190520e+07,  1.05084256e+08,
                -1.30219020e+07,  5.19906719e+05,  8.82430200e+06],
               [ 1.80755620e+07,  5.25578750e+06,  7.76211600e+07,
                -7.53063750e+06,  2.44717000e+06,  2.92675125e+06],
               [ 9.25567760e+07,  8.15244320e+07,  2.77157632e+08,
                -1.48489875e+06,  5.89250600e+07,  2.05608920e+07],
               [ 6.70443680e+07,  8.70343440e+07,  2.73050464e+08,
                -2.48670150e+06,  1.52268930e+07,  6.09583280e+07]])
        
        """
        request = field_pb2.GetElementaryDataRequest()
        request.field.CopyFrom(self._message)
        request.index = index
        list_message = self._stub.GetElementaryData(request, metadata=[(b'float_or_double', b'double')])
        data = []
        if list_message.elemdata_containers.data.HasField("datadouble"):
            data = list_message.elemdata_containers.data.datadouble.rep_double
        elif list_message.elemdata_containers.data.HasField("dataint"):
            data = list_message.elemdata_containers.data.dataint.rep_int

        array = np.array(data)
        if self.component_count !=1:
            n_comp = self.component_count
            array = array.reshape((len(data)//n_comp, n_comp))

        return array

    def get_entity_data_by_id(self, id):
        """Return the data of the scoping's id in parameter of the field.

        Returns
        -------
        data : numpy.array
            Data based on the scoping id.            
                  
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> stress_op = model.results.stress()
        >>> fields_container = stress_op.outputs.fields_container()
        >>> fields_container[0].get_entity_data_by_id(391)
        array([[-3.27795062e+05,  1.36012200e+06,  1.49090608e+08,
                -4.88688900e+06,  1.43038560e+07,  1.65455040e+07],
               [-4.63817550e+06,  1.29312225e+06,  1.20411832e+08,
                -6.06617800e+06,  2.34829700e+07,  1.77231120e+07],
               [-2.35684860e+07, -3.53474400e+07,  2.01501168e+08,
                -5.23361700e+06, -2.88789280e+07, -6.16478200e+06],
               [-3.92756960e+07, -2.72369280e+07,  1.81454016e+08,
                -3.75441450e+06, -3.62480300e+06, -3.26075620e+07],
               [ 1.63554530e+07,  2.83190520e+07,  1.05084256e+08,
                -1.30219020e+07,  5.19906719e+05,  8.82430200e+06],
               [ 1.80755620e+07,  5.25578750e+06,  7.76211600e+07,
                -7.53063750e+06,  2.44717000e+06,  2.92675125e+06],
               [ 9.25567760e+07,  8.15244320e+07,  2.77157632e+08,
                -1.48489875e+06,  5.89250600e+07,  2.05608920e+07],
               [ 6.70443680e+07,  8.70343440e+07,  2.73050464e+08,
                -2.48670150e+06,  1.52268930e+07,  6.09583280e+07]])
        
        """
        index = self.scoping.index(id)
        if index < 0:
            raise ValueError(f'The id {id} must be greater than 0')
        return self.get_entity_data(index)

    def append(self, data, scopingid):
        """add an entity data to the existing data

        Parameters
        ----------
        data : list of int, double or array

        scopingid : int
            id of the scoping
                
        Examples
        --------
        >>> from ansys.dpf.core import fields_factory
        >>> field = fields_factory.create_3d_vector_field(2)
        >>> field.append([1.,2.,3.],1)        
        >>> field.append([1.,2.,3.],2)
        >>> field.data
        array([[1., 2., 3.],
               [1., 2., 3.]])
        >>> field.scoping.ids
        [1, 2]
        
        """
        if isinstance(data, (np.ndarray, np.generic)):
            data = data.reshape(data.size).tolist()
        elif len(data)>0 and isinstance(data[0], list):
            data = np.array(data)
            data = data.reshape(data.size).tolist()
        request = field_pb2.AddDataRequest()
        if self._message.datatype == u"int":
            request.elemdata_containers.data.dataint.rep_int.extend(data)
        else:
            request.elemdata_containers.data.datadouble.rep_double.extend(data)
        request.elemdata_containers.scoping_id = scopingid

        request.field.CopyFrom(self._message)
        self._stub.AddData(request)
           
    @property
    def _data_pointer(self):
        """Gives the first index of each entity data

        Returns
        -------
        data : numpy.ndarray
            Data of this field.    
        
        Notes
        -----
        Print a progress bar
        """
        request = field_pb2.ListRequest()
        request.field.CopyFrom(self._message)
        service = self._stub.ListDataPointer(request)
        dtype = np.int32
        return scoping._data_get_chunk_(dtype, service)
    
    @property
    def _data_pointer_as_list(self):
        """Gives the first index of each entity data

        Returns
        -------
        data : list of int
            Data of this field.    
        
        Notes
        -----
        Print a progress bar
        """
        request = field_pb2.ListRequest()
        request.field.CopyFrom(self._message)
        service = self._stub.ListDataPointer(request)
        dtype = np.int32
        return scoping._data_get_chunk_(dtype, service, False)
    
    @_data_pointer.setter
    def _data_pointer(self, data):
        """Set the data pointer of the field.

        Parameters
        ----------
        data : list of int or array    
        
        Notes
        -----
        Print a progress bar
        """
        self._set_data_pointer(data)
        
           
    def _set_data_pointer(self,data):
        if isinstance(data,  (np.ndarray, np.generic)):
            data = np.array(data.reshape(data.size), dtype=np.int32)
        else:
            data = np.array(data, dtype=np.int32)
        if data.size ==0:
            return
        metadata=[(u"size_int", f"{len(data)}")]
        request = field_pb2.UpdateDataRequest()
        request.field.CopyFrom(self._message)
        self._stub.UpdateDataPointer(scoping._data_chunk_yielder(request, data), metadata=metadata)
        
        
    @property
    def data(self):
        """Access the data of this field.    
        
        Notes
        -----
        Print a progress bar
        
        Returns
        -------
        data : numpy.ndarray
            Data of this field.
        """
        return self._get_data()
    
    @property
    def data_as_list(self):
        """The data of this field in a python list

        Returns
        -------
        data : list
            Data of this field.    
        
        Notes
        -----
        Print a progress bar

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> fields_container = disp.outputs.fields_container()
        >>> field = fields_container[0]
        >>> # field.data_as_list
         
         """
        return self._get_data(np_array=False)
    
    def _get_data(self, np_array=True):
        request = field_pb2.ListRequest()
        request.field.CopyFrom(self._message)
        if self._message.datatype == u"int":
            data_type = u"int"
            dtype = np.int32
        else:
            data_type = u"double"
            dtype = np.float
        service = self._stub.List(request, metadata=[(u"float_or_double", data_type)])
        array= scoping._data_get_chunk_(dtype, service, np_array)
        
        ncomp = self.component_count
        if ncomp != 1 and np_array:
            array = array.reshape(self.shape)
        
        return array
    
    
    @data.setter
    def data(self, data):
        """Set the data of the field.    
        
        Notes
        -----
        Print a progress bar

        Parameters
        ----------
        data : list of int (property field only), double or array
        """
        self._set_data(data)
    
    def _set_data(self,data):
        if self._message.datatype == u"int":
            if not isinstance(data[0], int)and not isinstance(data[0], np.int32):
                raise errors.InvalidTypeError("data", "list of int")
            data = np.array(data, dtype=np.int32)
            metadata=[(u"size_int", f"{len(data)}")]
        else:
            if isinstance(data,  (np.ndarray, np.generic)):
                if 0 != self.size and self.component_count >1 and data.size//self.component_count != data.size/self.component_count:
                    raise ValueError(f'An array of shape {self.shape} is expected and shape {data.shape} is in input')
                else:
                    data = np.array(data.reshape(data.size), dtype=float)
            else:
                data = np.array(data, dtype=float)
            metadata=[(u"float_or_double", u"double"), (u"size_double", f"{len(data)}")]
        request = field_pb2.UpdateDataRequest()
        request.field.CopyFrom(self._message)
        self._stub.UpdateData(scoping._data_chunk_yielder(request, data), metadata=metadata)
        
    
    

class _LocalFieldBase(_FieldBase):
    """Class only created by a field to cache the internal data of the field,
    modify it locallly, and send a single update request to the server 
    when the local field is deleted
    
    Parameters
    ----------
    field : _FieldBase
        field to copy
    """
    def __init__(self, field):
        self._message = field._message
        self._server =field._server
        self._stub = field._stub
        self._is_property_field = field._message.datatype == u"int"
        self._owner_field = field
        self.__cache_data__()
        
    def __cache_data__(self):
        self._ncomp = super().component_count
        self._data_copy = super().data_as_list
        self._num_entities_reserved = len(self._data_copy)
        self._data_pointer_copy = super()._data_pointer_as_list
        self._scoping_ids_copy = super().scoping.ids
        self._num_entities = len(self._scoping_ids_copy)
        self._has_data_pointer = len(self._data_pointer_copy)>0
    
    @property
    def size(self):
        """The length of the data vector.
        Also equals to the number of elementary data times the number of components.
        
        Returns
        -------
        size : int
        """
        return len(self._data_copy)
    
    def get_entity_data(self, index):
        """Returns the elementary data of the scoping's index in parameter

        Returns
        -------
        data : numpy.array
        
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> stress_op = model.results.stress()
        >>> fields_container = stress_op.outputs.fields_container()
        >>> field = fields_container[0]
        >>> with field.as_local_field() as f:
        ...     print(f.get_entity_data(0))
        [[-3.27795062e+05  1.36012200e+06  1.49090608e+08 -4.88688900e+06
           1.43038560e+07  1.65455040e+07]
         [-4.63817550e+06  1.29312225e+06  1.20411832e+08 -6.06617800e+06
           2.34829700e+07  1.77231120e+07]
         [-2.35684860e+07 -3.53474400e+07  2.01501168e+08 -5.23361700e+06
          -2.88789280e+07 -6.16478200e+06]
         [-3.92756960e+07 -2.72369280e+07  1.81454016e+08 -3.75441450e+06
          -3.62480300e+06 -3.26075620e+07]
         [ 1.63554530e+07  2.83190520e+07  1.05084256e+08 -1.30219020e+07
           5.19906719e+05  8.82430200e+06]
         [ 1.80755620e+07  5.25578750e+06  7.76211600e+07 -7.53063750e+06
           2.44717000e+06  2.92675125e+06]
         [ 9.25567760e+07  8.15244320e+07  2.77157632e+08 -1.48489875e+06
           5.89250600e+07  2.05608920e+07]
         [ 6.70443680e+07  8.70343440e+07  2.73050464e+08 -2.48670150e+06
           1.52268930e+07  6.09583280e+07]]
        
        """
        if index > self._num_entities:
            raise ValueError(f"asked scoping {index} is greater than the number of available indices {len(self._scoping_ids_copy)}")
        if self._has_data_pointer:
            first_index = self._data_pointer_copy[index]
            if index < len(self._data_pointer_copy) -1:
                last_index =  self._data_pointer_copy[index+1]-1
            else:
                last_index = len(self._data_copy)-1
        else:
            first_index = self._ncomp * index
            last_index = self._ncomp * (index+1)-1
        if self._is_property_field:
            array = np.array(self._data_copy[first_index:last_index+1], dtype=np.int32)
        else:
            array = np.array(self._data_copy[first_index:last_index+1])
            
        if self._ncomp>1:
            return array.reshape((array.size//self._ncomp,self._ncomp))
        else:
            return array
    
    def get_entity_data_by_id(self, id):
        """Return the data of the scoping's id in parameter of the field.

        Returns
        -------
        data : numpy.array
            Data based on the scoping id.
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.static_rst)
        >>> stress_op = model.results.stress()
        >>> fields_container = stress_op.outputs.fields_container()
        >>> with fields_container[0].as_local_field() as f:
        ...     for id in f.scoping_ids:
        ...         if id < 2:
        ...             print(f.get_entity_data_by_id(id))
        [[-5.83890625e+03 -1.04498969e+05 -5.83890625e+03  2.10637354e+03
          -2.10637354e+03 -1.45397385e+02]
         [ 3.53322632e+03 -1.00388367e+05  3.53322632e+03 -1.66410352e+03
           1.66410352e+03  5.36620178e+01]
         [-1.05799683e+03 -1.00437922e+05  2.14961670e+03  5.90637268e+02
           1.37861340e+03 -1.68223175e+02]
         [ 2.62742480e+03 -9.89340078e+04 -2.02909998e+03 -2.40310791e+03
          -1.84942798e+03  7.16406616e+02]
         [-2.02909998e+03 -9.89340078e+04  2.62742480e+03  1.84942798e+03
           2.40310791e+03  7.16406616e+02]
         [ 2.14961670e+03 -1.00437922e+05 -1.05799683e+03 -1.37861340e+03
          -5.90637268e+02 -1.68223175e+02]
         [-4.94986755e+02 -9.87357891e+04 -4.94986755e+02 -6.93923187e+01
           6.93923187e+01 -1.59779755e+02]
         [ 1.36953296e+03 -9.76330156e+04  1.36953296e+03 -7.69014221e+02
           7.69014221e+02  4.90502930e+02]]

        """
        index = self._scoping_ids_copy.index(id)
        if index < 0:
            raise ValueError(f"The id {id} doesn't exist in the scoping")
        return self.get_entity_data(index)
    
    def append(self, data, scopingid):
        """Add an entity data to the existing data

        Parameters
        ----------
        data : list of int, double or array

        scopingid : int
            id of the scoping
            
        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> num_entities=100
        >>> field_to_local = dpf.fields_factory.create_3d_vector_field(num_entities, location=dpf.locations.elemental_nodal)
        >>> with field_to_local.as_local_field() as f:    
        ...     for i in range(1,num_entities+1):
        ...         f.append([[0.1*i,0.2*i, 0.3*i],[0.1*i,0.2*i, 0.3*i]],i)
                    
        """
        if self._is_property_field:
            if isinstance(data[0], np.int64):
                data = np.array(data, dtype=np.int32)
            if not isinstance(data[0], int) and not isinstance(data[0], np.int32):
                raise errors.InvalidTypeError("data", "list of int")
        if (len(data)>0 and isinstance(data, list)) or isinstance(data,  (np.ndarray, np.generic)):
                data=np.array(data).flatten().tolist()
            
        data_size =len(self._data_copy)      
        self._scoping_ids_copy.append(scopingid)
        if len(self._data_pointer_copy)>0:
            self._data_pointer_copy.append(data_size)

        self._data_copy.extend(data)
        self._num_entities+=1
        if self._has_data_pointer==False:
            if isinstance(data,  (np.ndarray, np.generic)):
                data_size = data.size
            else:
                data_size=len(data)
            if data_size>self._ncomp:
                self._data_pointer_copy=[i*self._ncomp for i in range(0,self._num_entities) ]
                self._has_data_pointer=True   
                
    def data_as_list(self):
        """The data of this field in a python list

        Returns
        -------
        data : list
            Data of this field.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> fields_container = disp.outputs.fields_container()
        >>> field = fields_container[0]
        >>> with field.as_local_field() as f:
        ...     my_data_list = f.data_as_list
         
        """
        return self._data_copy   
    
     
    @property
    def data(self):
        """The data of this field.

        Returns
        -------
        data : numpy.ndarray
            Data of this field.

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> transient = examples.download_transient_result()
        >>> model = dpf.Model(transient)
        >>> disp = model.results.displacement()
        >>> fields_container = disp.outputs.fields_container()
        >>> field = fields_container[0]
        >>> with field.as_local_field() as f:
        ...     print(f.data)
        [[ 6.25586668e-03 -1.39243136e-02  2.42697211e-05]
         [ 1.79675948e-02 -2.74812825e-02  1.83822050e-05]
         [-6.72664571e-03 -3.21373459e-02  1.67159110e-04]
         ...
         [-6.07730368e-03  3.22569017e-02  3.10184480e-04]
         [-3.51074714e-06  2.16872928e-08  6.40738989e-05]
         [ 1.03542516e-02 -3.53018374e-03 -3.98914380e-05]]
        
        """
        if self._ncomp>1:
            return np.array(self._data_copy).reshape(len(self._data_copy)//self._ncomp,self._ncomp)
        else:
            return np.array(self._data_copy)
        
    
    @data.setter
    def data(self, data):
        if self._is_property_field:
            if not isinstance(data[0], int)and not isinstance(data[0], np.int32):
                raise errors.InvalidTypeError("data", "list of int")
        else:
            if isinstance(data,  (np.ndarray, np.generic)):
                if data.shape !=  self.shape and 0 != self.size:
                    raise ValueError(f'An array of shape {self.shape} is expected and shape {data.shape} is in input')
        if isinstance(data,  (np.ndarray, np.generic)):
            self._data_copy = data.flatten().tolist()
        elif len(data)>0 and isinstance(data, list):
            self._data_copy=np.array(data).flatten().tolist()
        else:
            self._data_copy = data
        
        
    @property
    def elementary_data_count(self):
        """Number of elementary data in the field"""
        if (hasattr(self, "_data_copy")):
            return len(self._data_copy) / self._ncomp
        else:
            return super().elementary_data_count
    
    
    @property
    def component_count(self):
        """
        Returns
        -------
        ncomp : int
            Number of component of the each elementary data
        """
        return self._ncomp
    
        
    @property
    def _data_pointer(self):
        """Gives the first index of each entity data

        Returns
        -------
        data : numpy.ndarray
            Data of this field.
        """
        return np.array(self._data_pointer_copy)
    
    @property
    def _data_pointer_as_list(self):
        """Gives the first index of each entity data

        Returns
        -------
        data : list of int
            Data of this field.
        """
        return self._data_pointer_copy
    
    
    @_data_pointer.setter
    def _data_pointer(self, data):
        """Set the data pointer of the field.

        Parameters
        ----------
        data : list of int or array
        """
        if isinstance(data,  (np.ndarray, np.generic)):
            self._data_pointer_copy = data.tolist()
        else:
            self._data_pointer_copy = data
        if self._has_data_pointer == False and len(data)>0:
            self._has_data_pointer=True
    
    @property
    def scoping_ids(self):
        """List of int representing the scoping ids of the field.
        """
        return self._scoping_ids_copy
    
    @scoping_ids.setter
    def scoping_ids(self, data):
        self._scoping_ids_copy =data
        self._num_entities =len(data)
    
        
    def release_data(self):
        super()._set_data(self._data_copy)
        super()._set_data_pointer(self._data_pointer_copy)
        super().scoping.ids = self._scoping_ids_copy
        
    def __enter__(self):
        return self
    
    def __exit__(self,type, value, tb):
        if tb is None:
            self.release_data()
        else :
            print(tb)
            
    def __del__(self):
        pass
     

            
        


from grpc._channel import _InactiveRpcError, _MultiThreadedRendezvous
from functools import wraps

_COMPLEX_PLOTTING_ERROR_MSG = """
Complex fields can not be plotted. Use operators to get the amplitude
or the result at a defined sweeping phase before plotting.
"""

_FIELD_CONTAINER_PLOTTING_MSG = """"
This fields_container contains multiple fields.  Only one time-step
result can be plotted at a time.  Extract a field with
``fields_container[index]``.
"""

class DpfVersionNotSupported(RuntimeError):
    """Raised when the dpf-core/grpc-dpf python features are not 
    supported by DPF gRPC server version. """
    
    def __init__(self, version, msg = None):
        if msg is None: 
            msg = 'Feature not supported. Please upgrade the server to '
            msg += str(version)
            msg += ' version (or above).'
        RuntimeError.__init__(self, msg)

class DpfValueError(ValueError):
    """Raised when specific dpf error value must be defined"""
    
    def __init__(self, msg='A value that has been set leads to incorrect dpf behavior.'):
        ValueError.__init__(self, msg)
        
        
class InvalidTypeError(ValueError):
    """Raised when a parameter has the wrong type"""
    
    def __init__(self, data_type, parameter_name):
        msg = 'A ' + data_type + ' must be used for the following parameter: ' + parameter_name + '.'
        ValueError.__init__(self, msg)


class LocationError(ValueError):
    """Raised when using an invalid location"""

    def __init__(self, msg='Invalid location'):
        ValueError.__init__(self, msg)


class ComplexPlottingError(ValueError):
    """Raised when attempting to plot a field with complex data"""

    def __init__(self, msg=_COMPLEX_PLOTTING_ERROR_MSG):
        ValueError.__init__(self, msg)


class FieldContainerPlottingError(ValueError):
    """Raised when attempting to plot a fields_container containing
    multiple fields."""

    def __init__(self, msg=_FIELD_CONTAINER_PLOTTING_MSG):
        ValueError.__init__(self, msg)


class InvalidANSYSVersionError(RuntimeError):
    """Raised when ANSYS is an invalid version"""

    def __init__(self, msg=''):
        RuntimeError.__init__(self, msg)


class DPFServerException(Exception):
    """Raised when the DPF Server has encountered an error"""

    def __init__(self, msg=''):
        Exception.__init__(self, msg)


class DPFServerNullObject(Exception):
    """Raised when the DPF Server cannot find an object"""

    def __init__(self, msg=''):
        Exception.__init__(self, msg)


class InvalidPortError(OSError):
    """Raised when used an invalid port when starting DPF"""

    def __init__(self, msg=''):
        OSError.__init__(self, msg)


def protect_grpc(func):
    """Capture gRPC exceptions and return a more succinct error message"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        """Capture gRPC exceptions"""
        # Capture gRPC exceptions
        try:
            out = func(*args, **kwargs)
        except (_InactiveRpcError, _MultiThreadedRendezvous) as error:
            details = error.details()
            if 'object is null in the dataBase' in details:
                raise DPFServerNullObject(details) from None
            raise DPFServerException(details) from None            

        return out

    return wrapper



"""
.. _ref_operator:
    
Operator
===============
Interface to underlying gRPC Operator
"""
from textwrap import wrap
import logging
import grpc
import functools

from ansys.grpc.dpf import operator_pb2, operator_pb2_grpc, base_pb2
from ansys.dpf.core.inputs import Inputs, _Inputs, Input
from ansys.dpf.core.outputs import Outputs, _Outputs, Output
from ansys.dpf.core.errors import protect_grpc
from ansys.dpf.core.config import Config
from ansys.dpf.core.mapping_types import types
from ansys.dpf.core import server as serverlib

LOG = logging.getLogger(__name__)
LOG.setLevel('DEBUG')

class Operator:
    """A class used to represent an Operator which is an elementary
    operation.The Operator is the only object used to create and 
    transform the data. When the operator is evaluated, it will 
    process the input information to compute its output with respect 
    to its description.

    A list of existing operators can be asked through the "html_doc"
    operator.

    Parameters
    ----------
    name : str
        Name of the operator.  For example 'U'.

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.

    Examples
    --------
    Create an operator from the library of operators

    >>> from ansys.dpf import core as dpf
    >>> disp_oper = dpf.operators.result.displacement()

    Create an operator from a model

    >>> from ansys.dpf.core import Model 
    >>> from ansys.dpf.core import examples
    >>> model = Model(examples.static_rst)
    >>> disp_oper = model.results.displacement()
    
    """

    def __init__(self, name, config = None, server=None):
        """Initialize the operator with its name by connecting to a
        stub.
        """
        if server is None:
            server = serverlib._global_server()

        self._server = server

        self.name = name
        self._stub = self._connect()

        self._message = None
        self._description = None
        self._inputs = None
        self._outputs = None

        self.__send_init_request(config)
        
        # add dynamic inputs
        if len(self._message.spec.map_input_pin_spec) > 0 and self._inputs==None:
            self._inputs = Inputs(self._message.spec.map_input_pin_spec, self)
        if len(self._message.spec.map_output_pin_spec) != 0 and self._outputs==None:
            self._outputs = Outputs(self._message.spec.map_output_pin_spec, self)
        
        self._description = self._message.spec.description

    def _add_sub_res_operators(self, sub_results):
        """Dynamically add operators instantiating for sub-results.

        The new operators subresults are connected to the parent
        operator's inputs when created, but are, then, completely
        independent of the parent operators.

        Examples
        --------
        >>> disp_oper = model.results.displacement()
        >>> disp_x = model.results.displacement().X()
        >>> disp_y = model.results.displacement().Y()
        >>> disp_z = model.results.displacement().Z()
        
        """
        for result_type in sub_results:             
            bound_method = self._sub_result_op.__get__(self, self.__class__)
            method2 = functools.partial(bound_method, name=result_type["operator name"])
            setattr(self, result_type["name"], method2)

        
    @protect_grpc
    def connect(self, pin, inpt, pin_out=0):
        """Connect an input on the operator using a pin number.

        Parameters
        ----------
        pin : int
            Number of the input pin.

        inpt :  str, int, double, bool, list of int, list of doubles, Field, FieldsContainer, Scoping, ScopingsContainer, MeshedRegion, MeshesContainer, DataSources, Operator
            Object you wish to connect.

        pin_out : int, optional
            In case of the input is an Operator, this is the output
            pin of the input Operator.  Defaults to 0.

        Examples
        --------
        Compute the minimum of displacement by chaining the ``'U'``
        and ``'min_max_fc'`` operators.

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        >>> max_fc_op = dpf.operators.min_max.min_max_fc()
        >>> max_fc_op.inputs.connect(disp_op.outputs)
        >>> max_field = max_fc_op.outputs.field_max()
        >>> max_field.data
        array([[0.59428386, 0.00201751, 0.0006032 ]])
        
        """
        request = operator_pb2.UpdateRequest()
        request.op.CopyFrom(self._message)        
        request.pin = pin
        _fillConnectionRequestMessage(request, inpt, pin_out)
        if inpt is self:
            raise ValueError('Cannot connect to itself')
        self._stub.Update(request)
    
        
    @protect_grpc
    def get_output(self, pin=0, output_type=None):
        """Returns the output of the operator on the pin number.

        Parameters
        ----------
        pin : int, optional
            Number of the output pin.

        output_type : core.type enum, optional
            The requested type of the output.
        """
        request = operator_pb2.OperatorEvaluationRequest()
        request.op.CopyFrom(self._message)
        request.pin = pin
        
        if output_type is not None:
            _write_output_type_to_proto_style(output_type, request)
            out = self._stub.Get(request)
            return _convertOutputMessageToPythonInstance(out, output_type, self._server)
        else:
            request.type = base_pb2.Type.Value('RUN')
            return self._stub.Get(request)
            
    
    @property
    def config(self):
        """Returns a copy of the current config of the operator.
        To use the config that you modify, please use operator.config = new_config
        or create an operator with the new config as a parameter.
        
        Returns
        ----------
        config : Config        
        """
        out = self._stub.List(self._message)
        config = out.config
        return Config(config =config, server = self._server)
    
    
    @config.setter
    def config(self,value):
        """Change the configuration of the operator 
        (if the operator is up to date, changing the config doesn't make it not up to date)
         
        Parameters
        ----------
        value : Config
        """
        request = operator_pb2.UpdateConfigRequest()
        request.op.CopyFrom(self._message)
        request.config.CopyFrom(value._message)
        self._stub.UpdateConfig(request)
        
    @property
    def inputs(self):
        """Enables to connect inputs to the operator 
        
        Returns
        --------
        inputs : Inputs
        
        Examples
        --------
        Use the displacement operator

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        
        """
        return self._inputs
    
    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it
        
        Returns
        --------
        outputs : Output
        
        Examples
        --------
        CUse the displacement operator

        >>> from ansys.dpf import core as dpf
        >>> from ansys.dpf.core import examples
        >>> data_src = dpf.DataSources(examples.multishells_rst)
        >>> disp_op = dpf.operators.result.displacement()
        >>> disp_op.inputs.data_sources(data_src)
        >>> disp_fc = disp_op.outputs.fields_container()
        
        """
        return self._outputs
        
      
    @staticmethod
    def default_config(name, server=None):
        """Returns the default config for a given operator.
        This config can then be changed to the user needs and be used to
        instantiate the given operator
        
        Parameters
        ----------
        name : str
            Name of the operator.  For example 'U'.
    
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the the global server.
        """
        return Config(operator_name = name, server =server)
        
    def _connect(self):
        """Connect to the grpc service"""
        return operator_pb2_grpc.OperatorServiceStub(self._server.channel)

    def __del__(self):
        try:
            self._stub.Delete(self._message)
        except:
            pass

    def __str__(self):
        """describe the entity
        
        Returns
        -------
        description : str
        """
        from ansys.dpf.core.core import _description
        return _description(self._message, self._server)

    def run(self):
        """Evaluate this operator"""
        self.get_output()

    def _find_outputs_corresponding_pins(self, type_names, inpt, pin,
                                         corresponding_pins):
        for python_name in type_names:
            # appears to be an issue on Linux.  This check is here
            # because cpp mappings are a single type mapping and
            # sometimes the spec contains 'B' instead of 'bool'
            if python_name == 'B':
                python_name = 'bool'

            if  type(inpt).__name__ == python_name:
                corresponding_pins.append(pin)
            elif isinstance(inpt, _Outputs) or isinstance(inpt,Operator):
                if isinstance(inpt,Operator):
                    output_pin_available = inpt.outputs._get_given_output([python_name])
                else:
                    
                    output_pin_available = inpt._get_given_output([python_name])
                for outputpin in output_pin_available:
                    corresponding_pins.append((pin, outputpin))
            elif isinstance(inpt, Output):
                for inpttype in inpt._python_expected_types:
                    if inpttype == python_name:
                        corresponding_pins.append(pin)
                if python_name == "Any":
                    corresponding_pins.append(pin)
            elif python_name == "Any":
                corresponding_pins.append(pin)

    @protect_grpc
    def _sub_result_op(self, name):
        op = Operator(name)
        if self.inputs is not None:
            for key in self.inputs._connected_inputs:
                inpt = self.inputs._connected_inputs[key]
                if type(inpt).__name__ == 'dict':
                    for keyout in inpt:
                        op.connect(key,inpt[keyout],keyout)
                else:
                    op.connect(key,inpt)
        return op

    @protect_grpc
    def __send_init_request(self, config = None):
        request = operator_pb2.CreateOperatorRequest()
        request.name = self.name
        if config:
            request.config.CopyFrom(config._message)
        self._message = self._stub.Create(request)

    def __add__(self, fields_b):
        """Adds two fields or fields containers together
                
        Returns
        -------
        add : operators.math.add_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "add_fc") :
            op= operators.math.add_fc(self, fields_b)
        else :
            op= dpf_operator.Operator("add_fc")
            op.connect(0,self)        
            op.connect(1, fields_b)
        return op
    
    
    def __sub__(self, fields_b):
        """Subtract two fields or fields containers together
                
        Returns
        -------
        minus : operators.math.minus_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "minus_fc") :
            op= operators.math.minus_fc()
        else :
            op= dpf_operator.Operator("minus_fc")
        op.connect(0,self)        
        op.connect(1, fields_b)
        return op
    

    def __pow__(self, value):
        if value != 2:
            raise ValueError('DPF only the value is "2" suppported')
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "sqr_fc") :
            op= operators.math.sqr_fc()
        else :
            op= dpf_operator.Operator("sqr_fc")
        op.connect(0,self)        
        op.connect(1, value)
        return op
    
    def __mul__(self, value):
        """Multiplies two fields or fields containers together
        
        Returns
        -------
        mul : operators.math.generalized_inner_product_fc
        """
        from ansys.dpf.core import dpf_operator
        from ansys.dpf.core import operators
        if hasattr(operators, "math") and  hasattr(operators.math, "generalized_inner_product_fc") :
            op= operators.math.generalized_inner_product_fc()
        else :
            op= dpf_operator.Operator("generalized_inner_product_fc")
        op.connect(0,self)        
        op.connect(1, value)
        return op
    
    

    def __truediv__(self, inpt):
        if isinstance(inpt, Operator):
            op = Operator("div")
            op.connect(0, self, 0)
            op.connect(1, inpt, 0)
        elif isinstance(inpt, float):
            op = Operator("scale")
            op.connect(0, self, 0)
            op.connect(1, 1.0/inpt)
        return op
    


def _write_output_type_to_proto_style(output_type, request):
    subtype=''
    stype=''
    if hasattr(output_type, 'name'):
        if output_type == types.fields_container:
            stype='collection'
            subtype = 'field'
        elif output_type== types.scopings_container:
            stype='collection'
            subtype = 'scoping'
        elif output_type== types.meshes_container:
            stype='collection'
            subtype = 'meshed_region'
        else :
            stype = output_type.name
    elif isinstance(output_type,list):
        stype=output_type[0]
        subtype = output_type[1]
    else:
        stype = output_type
    request.type = base_pb2.Type.Value(stype.upper())
    if subtype !="":
        request.subtype = base_pb2.Type.Value(subtype.upper())
            

def _convertOutputMessageToPythonInstance(out, output_type, server):
    from ansys.dpf.core import (fields_container, field, property_field, field_base, scopings_container, scoping,
                            meshes_container, meshed_region, result_info, time_freq_support, collection, data_sources,
                            collection, data_sources, cyclic_support)
    if out.HasField("str"):
        return out.str
    elif out.HasField("int"):
        return out.int
    elif out.HasField("double"):
        return out.double 
    elif out.HasField("bool"):
        return out.bool
    elif out.HasField("field"):
        toconvert = out.field
        if toconvert.datatype == u"int":
            return property_field.PropertyField(server=server,property_field=toconvert)
        else:   
            return field.Field(server=server,field=toconvert)
    elif out.HasField("collection"):
        toconvert = out.collection
        if output_type == types.fields_container:
            return fields_container.FieldsContainer(server=server,fields_container=toconvert)
        elif output_type == types.scopings_container:
            return scopings_container.ScopingsContainer(server=server,scopings_container=toconvert)
        elif output_type == types.meshes_container:
            return meshes_container.MeshesContainer(server=server,meshes_container=toconvert)
    elif out.HasField("scoping"):
        toconvert = out.scoping
        return scoping.Scoping(scoping=toconvert,server=server)
    elif out.HasField("mesh"):
        toconvert = out.mesh
        return meshed_region.MeshedRegion(mesh=toconvert, server=server)
    elif out.HasField("result_info"):
        toconvert = out.result_info
        return result_info.ResultInfo(result_info=toconvert, server=server)
    elif out.HasField("time_freq_support"):
        toconvert = out.time_freq_support
        return time_freq_support.TimeFreqSupport(server=server, time_freq_support=toconvert)
    elif out.HasField("data_sources"):
        toconvert = out.data_sources
        return data_sources.DataSources(server=server,data_sources=toconvert)  
    elif out.HasField("cyc_support"):
        toconvert = out.cyc_support
        return cyclic_support.CyclicSupport(server=server,cyclic_support=toconvert)
    
def _fillConnectionRequestMessage(request, inpt, pin_out=0):
    from ansys.dpf.core import (fields_container, field, property_field, field_base, scopings_container, scoping,
                            meshes_container, meshed_region, result_info, time_freq_support, collection, data_sources,
                            collection, data_sources, cyclic_support, model)
    if isinstance(inpt, str):
        request.str = inpt
    elif isinstance(inpt, bool):
        request.bool = inpt
    elif isinstance(inpt, int):
        request.int = inpt
    elif isinstance(inpt, float):
        request.double = inpt
    elif isinstance(inpt, list):
        if all(isinstance(x, int) for x in inpt):
            request.vint.rep_int.extend(inpt)
        elif all(isinstance(x, float) for x in inpt):
            request.vdouble.rep_double.extend(inpt)
    elif isinstance(inpt, field_base._FieldBase):
        request.field.CopyFrom(inpt._message)
    elif isinstance(inpt, collection.Collection):
        request.collection.CopyFrom(inpt._message)
    elif isinstance(inpt, scoping.Scoping):
        request.scoping.CopyFrom(inpt._message)
    elif isinstance(inpt, data_sources.DataSources):
        request.data_sources.CopyFrom(inpt._message)      
    elif isinstance(inpt, model.Model):
        request.data_sources.CopyFrom(inpt.metadata.data_sources._message)
    elif isinstance(inpt, meshed_region.MeshedRegion):
        request.mesh.CopyFrom(inpt._message)
    elif isinstance(inpt, cyclic_support.CyclicSupport):
        request.cyc_support.CopyFrom(inpt._message)
    elif isinstance(inpt, Operator):
        request.inputop.inputop.CopyFrom(inpt._message)
        request.inputop.pinOut = pin_out
    elif isinstance(inpt, Output):
        request.inputop.inputop.CopyFrom(inpt._operator._message)
        request.inputop.pinOut = inpt._pin
    else:
        errormsg = f"input type {inpt.__class__} cannot be connected"
        raise TypeError(errormsg)
        
    


"""
Custom FieldsContainers
=======================
Contains inherited class of the FieldsContainer.
Those new classes offer helpers to access data for specific usage:
results split by body, by material...
"""
from ansys.dpf.core.fields_container import FieldsContainer
from ansys.grpc.dpf import meshed_region_pb2


class ElShapeFieldsContainer(FieldsContainer):
    """A class used to represent a FieldsContainer with fields split by
    element shapes : solid, shell, beam...
    Instances of this class are created by a result of the model asked to be 
    split by element shape

    Parameters
    ----------
    fields_container : ansys.grpc.dpf.collection_pb2.Collection or FieldsContainer, optional
        Create a fields container from a Collection message or create a copy from an existing fields container

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
    
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.download_all_kinds_of_complexity_modal())
    >>> fc = model.results.displacement.on_all_time_freqs.split_by_shape.eval()
    >>> len(fc.solid_fields())
    45
    >>> solid_f_time_2 = fc.solid_field(2)
    
    """
    
    def __init__(self, fields_container=None, server=None):
        super().__init__(fields_container, server)
        if not fields_container:
            self.add_label("elshape")
    
    def solid_fields(self, timeid=None, complexid=None):
        """Returns a list of all the fields with solid element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the list of fields returned are solid fields for a given time and/or 
        complex type
        
        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
        
        Examples
        --------
        >>> len(fc.solid_fields())
        45
        >>> len(fc.solid_fields(timeid=3))
        1
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("SOLID")
        return self.get_fields(label_space)
    
    def shell_fields(self, timeid=None, complexid=None):
        """Returns a list of all the fields with shell element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the list of fields returned are shell fields for a given time and/or 
        complex type
        
        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
        
        Examples
        --------
        >>> len(fc.shell_fields())
        45
        >>> len(fc.shell_fields(timeid=3))
        1
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("SHELL")
        return self.get_fields(label_space)
    
    def beam_fields(self, timeid=None, complexid=None):
        """Returns a list of all the fields with beam element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the list of fields returned are beam fields for a given time and/or 
        complex type
        
        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
        
        Examples
        --------
        >>> len(fc.beam_fields())
        45
        >>> len(fc.beam_fields(timeid=3))
        1
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("BEAM")
        return self.get_fields(label_space)
    
    def solid_field(self, timeid=None, complexid=None):
        """Returns the requested field with solid element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the field returned is a solid field for a given time and/or 
        complex type. It throws if the number of fields matching the request is 
        higher than 1.
        
        Returns
        -------
        fields: Field
            field corresponding to the request
        
        Examples
        --------
        >>> field = fc.solid_field(timeid=3)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("SOLID")
        return self.get_field(label_space)
    
    def shell_field(self, timeid=None, complexid=None):
        """Returns the requested field with shell element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the field returned is a shell field for a given time and/or 
        complex type. It throws if the number of fields matching the request is 
        higher than 1.
        
        Returns
        -------
        fields: Field
            field corresponding to the request
        
        Examples
        --------
        >>> field = fc.shell_field(timeid=3)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("SHELL")
        return self.get_field(label_space)
    
    def beam_field(self, timeid=None, complexid=None):
        """Returns the requested field with beam element shapes.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the field returned is a beam field for a given time and/or 
        complex type. It throws if the number of fields matching the request is 
        higher than 1.
        
        Returns
        -------
        fields: Field
            field corresponding to the request
        
        Examples
        --------
        >>> field = fc.beam_field(timeid=3)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["elshape"]= meshed_region_pb2.ElementShape.Value("BEAM")
        return self.get_field(label_space)
    

class BodyFieldsContainer(FieldsContainer):
    """A class used to represent a FieldsContainer with fields split by
    body (mapdl material property)
    Instances of this class are created by a result of the model asked to be 
    split by body

    Parameters
    ----------
    fields_container : ansys.grpc.dpf.collection_pb2.Collection or FieldsContainer, optional
        Create a fields container from a Collection message or create a copy from an existing fields container

    server : server.DPFServer, optional
        Server with channel connected to the remote or local instance. When
        ``None``, attempts to use the the global server.
    
    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.download_all_kinds_of_complexity_modal())
    >>> fc = model.results.displacement.on_all_time_freqs.split_by_body.eval()
    >>> fc.get_mat_scoping().ids[3]
    45
    >>> len(fc.get_fields_by_mat_id(45))
    45
    >>> f_time_2 = fc.get_field_by_mat_id(45, timeid=2)
    
    """
    
    def __init__(self, fields_container=None, server=None):
        super().__init__(fields_container, server)
        if not fields_container:
            self.add_label("elshape")
    
    def get_fields_by_mat_id(self, matid, timeid=None, complexid=None):
        """Returns a list of all the fields with the given material id.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the list of fields returned are fields for a given time and/or 
        complex type. The available mat id can be requested with `get_mat_scoping`
        
        Returns
        -------
        fields : list[Field]
            fields corresponding to the request
        
        Examples
        --------
        >>> len(fc.get_fields_by_mat_id(1))
        45
        >>> len(fc.get_fields_by_mat_id(1, timeid=3))
        1
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["mat"]= matid
        return self.get_fields(label_space)
    
    def get_field_by_mat_id(self, matid, timeid=None, complexid=None):
        """Returns a a field with the given material id.
        If a timeid or a complexid (0 for real and 1 for imaginary) are given, 
        the field returned is for a given time and/or 
        complex type. The available mat id can be requested with `get_mat_scoping`
        
        Returns
        -------
        fields : Field
            field corresponding to the request
        
        Examples
        --------
        >>> f_time_2 = fc.get_field_by_mat_id(45, timeid=2)
        
        """
        label_space = self.__time_complex_label_space__(timeid,complexid)
        label_space["mat"]= matid
        return self.get_field(label_space)
    
    def get_mat_scoping(self):
        """Returns the material or body scoping containing the mat ids
        
        Returns
        -------
        scoping: Scoping
            scoping containing the mat ids available in the fields container
        """
        return self.get_label_scoping("mat")


"""
scoping_on_coordinates
======================
"""
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs, _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification

"""Operators from Ans.Dpf.FEMutils plugin, from "mapping" category
"""

class scoping_on_coordinates(Operator):
    """Finds the Elemental scoping of a set of coordinates.

      available inputs:
        - coordinates (Field)
        - mesh (MeshedRegion)

      available outputs:
        - scoping (Scoping)

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> # Instantiate operator
      >>> op = dpf.operators.mapping.scoping_on_coordinates()

      >>> # Make input connections
      >>> my_coordinates = dpf.Field()
      >>> op.inputs.coordinates.connect(my_coordinates)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)

      >>> # Instantiate operator and connect inputs in one line
      >>> op = dpf.operators.mapping.scoping_on_coordinates(coordinates=my_coordinates,mesh=my_mesh)

      >>> # Get output data
      >>> result_scoping = op.outputs.scoping()"""
    def __init__(self, coordinates=None, mesh=None, config=None, server=None):
        super().__init__(name="scoping::on_coordinates", config = config, server = server)
        self._inputs = InputsScopingOnCoordinates(self)
        self._outputs = OutputsScopingOnCoordinates(self)
        if coordinates !=None:
            self.inputs.coordinates.connect(coordinates)
        if mesh !=None:
            self.inputs.mesh.connect(mesh)

    @staticmethod
    def _spec():
        spec = Specification(description="""Finds the Elemental scoping of a set of coordinates.""",
                             map_input_pin_spec={
                                 0 : PinSpecification(name = "coordinates", type_names=["field"], optional=False, document=""""""), 
                                 7 : PinSpecification(name = "mesh", type_names=["abstract_meshed_region"], optional=False, document="""""")},
                             map_output_pin_spec={
                                 0 : PinSpecification(name = "scoping", type_names=["scoping"], optional=False, document="""""")})
        return spec


    @staticmethod
    def default_config():
        return Operator.default_config(name = "scoping::on_coordinates")

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsScopingOnCoordinates 
        """
        return super().inputs


    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluationg it

        Returns
        --------
        outputs : OutputsScopingOnCoordinates 
        """
        return super().outputs


#internal name: scoping::on_coordinates
#scripting name: scoping_on_coordinates
class InputsScopingOnCoordinates(_Inputs):
    """Intermediate class used to connect user inputs to scoping_on_coordinates operator

      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mapping.scoping_on_coordinates()
      >>> my_coordinates = dpf.Field()
      >>> op.inputs.coordinates.connect(my_coordinates)
      >>> my_mesh = dpf.MeshedRegion()
      >>> op.inputs.mesh.connect(my_mesh)
    """
    def __init__(self, op: Operator):
        super().__init__(scoping_on_coordinates._spec().inputs, op)
        self._coordinates = Input(scoping_on_coordinates._spec().input_pin(0), 0, op, -1) 
        self._inputs.append(self._coordinates)
        self._mesh = Input(scoping_on_coordinates._spec().input_pin(7), 7, op, -1) 
        self._inputs.append(self._mesh)

    @property
    def coordinates(self):
        """Allows to connect coordinates input to the operator

        Parameters
        ----------
        my_coordinates : Field, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mapping.scoping_on_coordinates()
        >>> op.inputs.coordinates.connect(my_coordinates)
        >>> #or
        >>> op.inputs.coordinates(my_coordinates)

        """
        return self._coordinates

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator

        Parameters
        ----------
        my_mesh : MeshedRegion, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mapping.scoping_on_coordinates()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> #or
        >>> op.inputs.mesh(my_mesh)

        """
        return self._mesh

class OutputsScopingOnCoordinates(_Outputs):
    """Intermediate class used to get outputs from scoping_on_coordinates operator
      Examples
      --------
      >>> from ansys.dpf import core as dpf

      >>> op = dpf.operators.mapping.scoping_on_coordinates()
      >>> # Connect inputs : op.inputs. ...
      >>> result_scoping = op.outputs.scoping()
    """
    def __init__(self, op: Operator):
        super().__init__(scoping_on_coordinates._spec().outputs, op)
        self._scoping = Output(scoping_on_coordinates._spec().output_pin(0), 0, op) 
        self._outputs.append(self._scoping)

    @property
    def scoping(self):
        """Allows to get scoping output of the operator


        Returns
        ----------
        my_scoping : Scoping, 

        Examples
        --------
        >>> from ansys.dpf import core as dpf

        >>> op = dpf.operators.mapping.scoping_on_coordinates()
        >>> # Connect inputs : op.inputs. ...
        >>> result_scoping = op.outputs.scoping() 
        """
        return self._scoping

